import numpy as np
from astropy.coordinates import SkyCoord
from astropy_healpix import HEALPix, npix_to_nside

from reproject._array_utils import map_coordinates

__all__ = ["healpix_to_image", "image_to_healpix"]

ORDER = {}
ORDER["nearest-neighbor"] = 0
ORDER["bilinear"] = 1
ORDER["biquadratic"] = 2
ORDER["bicubic"] = 3


def healpix_to_image(
    healpix_data, coord_system_in, wcs_out, shape_out, order="bilinear", nested=False
):
    """
    Convert image in HEALPIX format to a normal FITS projection image (e.g.
    CAR or AIT).

    Parameters
    ----------
    healpix_data : `numpy.ndarray`
        HEALPIX data array
    coord_system_in : str or `~astropy.coordinates.BaseCoordinateFrame`
        The coordinate system for the input HEALPIX data, as an Astropy
        coordinate frame or corresponding string alias (e.g. ``'icrs'`` or
        ``'galactic'``)
    wcs_out : `~astropy.wcs.WCS`
        The WCS of the output array
    shape_out : tuple
        The shape of the output array
    order : int or str, optional
        The order of the interpolation (if ``mode`` is set to
        ``'interpolation'``). This can be either one of the following strings:

            * 'nearest-neighbor'
            * 'bilinear'

        or an integer. A value of ``0`` indicates nearest neighbor
        interpolation.
    nested : bool
        The order of the healpix_data, either nested or ring.  Stored in
        FITS headers in the ORDERING keyword.

    Returns
    -------
    reprojected_data : `numpy.ndarray`
        HEALPIX image resampled onto the reference image
    footprint : `~numpy.ndarray`
        Footprint of the input array in the output array. Values of 0 indicate
        no coverage or valid values in the input image, while values of 1
        indicate valid values.
    """

    healpix_data = np.asarray(healpix_data, dtype=float)

    # Look up lon, lat of pixels in reference system and convert celestial coordinates
    yinds, xinds = np.indices(shape_out)
    world_in = wcs_out.pixel_to_world(xinds, yinds).transform_to(coord_system_in)
    world_in_unitsph = world_in.represent_as("unitspherical")
    lon_in, lat_in = world_in_unitsph.lon, world_in_unitsph.lat

    if isinstance(order, str):
        order = ORDER[order]

    nside = npix_to_nside(len(healpix_data))

    hp = HEALPix(nside=nside, order="nested" if nested else "ring")

    if order == 1:
        with np.errstate(invalid="ignore"):
            data = hp.interpolate_bilinear_lonlat(lon_in, lat_in, healpix_data)
        footprint = (~np.isnan(data)).astype(float)
    elif order == 0:
        with np.errstate(invalid="ignore"):
            ipix = hp.lonlat_to_healpix(lon_in, lat_in)
        data = healpix_data[ipix]
        footprint = (ipix != -1).astype(float)
    else:
        raise ValueError("Only nearest-neighbor and bilinear interpolation are supported")

    return data, footprint


def image_to_healpix(data, wcs_in, coord_system_out, nside, order="bilinear", nested=False):
    """
    Convert image in a normal WCS projection to HEALPIX format.

    Parameters
    ----------
    data : `numpy.ndarray`
        Input data array to reproject
    wcs_in : `~astropy.wcs.WCS`
        The WCS of the input array
    coord_system_out : str or `~astropy.coordinates.BaseCoordinateFrame`
        The target coordinate system for the HEALPIX projection, as an Astropy
        coordinate frame or corresponding string alias (e.g. ``'icrs'`` or
        ``'galactic'``)
    order : int or str, optional
        The order of the interpolation (if ``mode`` is set to
        ``'interpolation'``). This can be either one of the following strings:

            * 'nearest-neighbor'
            * 'bilinear'
            * 'biquadratic'
            * 'bicubic'

        or an integer. A value of ``0`` indicates nearest neighbor
        interpolation.
    nested : bool
        The order of the healpix_data, either nested or ring.  Stored in
        FITS headers in the ORDERING keyword.

    Returns
    -------
    reprojected_data : `numpy.ndarray`
        A HEALPIX array of values
    footprint : `~numpy.ndarray`
        Footprint of the input array in the output array. Values of 0 indicate
        no coverage or valid values in the input image, while values of 1
        indicate valid values.
    """

    hp = HEALPix(nside=nside, order="nested" if nested else "ring")

    npix = hp.npix

    # Look up lon, lat of pixels in output system and convert colatitude theta
    # and longitude phi to longitude and latitude.
    lon_out, lat_out = hp.healpix_to_lonlat(np.arange(npix))

    world_out = SkyCoord(lon_out, lat_out, frame=coord_system_out)

    # Look up pixels in input WCS
    xinds, yinds = wcs_in.world_to_pixel(world_out)

    # Interpolate

    if isinstance(order, str):
        order = ORDER[order]

    healpix_data = map_coordinates(
        data,
        np.array([yinds, xinds]),
        order=order,
        mode="constant",
        cval=np.nan,
    )

    return healpix_data, (~np.isnan(healpix_data)).astype(float)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import warnings

from astropy.utils.exceptions import AstropyDeprecationWarning

warnings.warn(
    "reproject.interpolation.high_level is a private module and will be removed in a future "
    "version.  Please use the public API from reproject.interpolation instead.",
    AstropyDeprecationWarning,
    stacklevel=2,
)

from ._high_level import *  # noqa: E402, F401, F403



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import itertools
import os

import numpy as np
import pytest
from astropy.io import fits
from astropy.wcs import WCS
from astropy_healpix import nside_to_npix

from ...interpolation.tests.test_core import as_high_level_wcs
from ...tests.test_high_level import ALL_DTYPES
from .._high_level import reproject_from_healpix, reproject_to_healpix
from .._utils import parse_coord_system

DATA = os.path.join(os.path.dirname(__file__), "data")


def get_reference_header(overscan=1, oversample=2, nside=1):
    reference_header = fits.Header()
    reference_header.update(
        {
            "CDELT1": -180.0 / (oversample * 4 * nside),
            "CDELT2": 180.0 / (oversample * 4 * nside),
            "CRPIX1": overscan * oversample * 4 * nside,
            "CRPIX2": overscan * oversample * 2 * nside,
            "CRVAL1": 180.0,
            "CRVAL2": 0.0,
            "CTYPE1": "RA---CAR",
            "CTYPE2": "DEC--CAR",
            "CUNIT1": "deg",
            "CUNIT2": "deg",
            "NAXIS": 2,
            "NAXIS1": overscan * oversample * 8 * nside,
            "NAXIS2": overscan * oversample * 4 * nside,
        }
    )

    return reference_header


@pytest.mark.parametrize(
    "nside,nested,healpix_system,image_system,dtype,order",
    list(
        itertools.product(
            [1, 2, 4, 8, 16, 32, 64],
            [True, False],
            ["C"],
            ["C"],
            ALL_DTYPES,
            ["bilinear", "nearest-neighbor"],
        )
    ),
)
def test_reproject_healpix_to_image_footprint(
    nside, nested, healpix_system, image_system, dtype, order
):
    """Test that HEALPix->WCS conversion correctly flags pixels that do not
    have valid WCS coordinates."""

    npix = nside_to_npix(nside)
    healpix_data = np.random.uniform(size=npix).astype(dtype)

    reference_header = get_reference_header(overscan=2, oversample=2, nside=nside)

    wcs_out = WCS(reference_header)
    shape_out = reference_header["NAXIS2"], reference_header["NAXIS1"]

    image_data, footprint = reproject_from_healpix(
        (healpix_data, healpix_system),
        wcs_out,
        shape_out=shape_out,
        order=order,
        nested=nested,
    )

    if order == "bilinear":
        expected_footprint = ~np.isnan(image_data)
    else:
        coord_system_in = parse_coord_system(healpix_system)
        yinds, xinds = np.indices(shape_out)
        world_in = wcs_out.pixel_to_world(xinds, yinds).transform_to(coord_system_in)
        world_in_unitsph = world_in.represent_as("unitspherical")
        lon_in, lat_in = world_in_unitsph.lon, world_in_unitsph.lat
        expected_footprint = ~(np.isnan(lon_in) | np.isnan(lat_in))

    np.testing.assert_array_equal(footprint, expected_footprint)


@pytest.mark.parametrize(
    "wcsapi,nside,nested,healpix_system,image_system,dtype",
    list(
        itertools.product(
            [True, False], [1, 2, 4, 8, 16, 32, 64], [True, False], ["C"], ["C"], ALL_DTYPES
        )
    ),
)
def test_reproject_healpix_to_image_round_trip(
    wcsapi, nside, nested, healpix_system, image_system, dtype
):
    """Test round-trip HEALPix->WCS->HEALPix conversion for a random map,
    with a WCS projection large enough to store each HEALPix pixel"""

    npix = nside_to_npix(nside)
    healpix_data = np.random.uniform(size=npix).astype(dtype)

    reference_header = get_reference_header(oversample=2, nside=nside)

    wcs_out = WCS(reference_header)
    shape_out = reference_header["NAXIS2"], reference_header["NAXIS1"]

    if wcsapi:
        wcs_out = as_high_level_wcs(wcs_out)

    image_data, footprint = reproject_from_healpix(
        (healpix_data, healpix_system),
        wcs_out,
        shape_out=shape_out,
        order="nearest-neighbor",
        nested=nested,
    )

    healpix_data_2, footprint = reproject_to_healpix(
        (image_data, wcs_out), healpix_system, nside=nside, order="nearest-neighbor", nested=nested
    )

    np.testing.assert_array_equal(healpix_data, healpix_data_2)


def test_reproject_file():
    reference_header = get_reference_header(oversample=2, nside=8)
    data, footprint = reproject_from_healpix(
        os.path.join(DATA, "bayestar.fits.gz"), reference_header
    )
    reference_result = fits.getdata(os.path.join(DATA, "reference_result.fits"))
    np.testing.assert_allclose(data, reference_result, rtol=1.0e-5)


def test_reproject_invalid_order():
    reference_header = get_reference_header(oversample=2, nside=8)
    with pytest.raises(ValueError) as exc:
        reproject_from_healpix(
            os.path.join(DATA, "bayestar.fits.gz"), reference_header, order="bicubic"
        )
    assert exc.value.args[0] == "Only nearest-neighbor and bilinear interpolation are supported"


def test_reproject_to_healpix_input_types(valid_celestial_input_data):
    array_ref, wcs_in_ref, input_value, kwargs_in = valid_celestial_input_data

    # For now we don't support 3D arrays in reproject_to_healpix
    if array_ref.ndim == 3:
        pytest.skip()

    # Compute reference

    healpix_data_ref, footprint_ref = reproject_to_healpix((array_ref, wcs_in_ref), "C", nside=64)

    # Compute test

    healpix_data_test, footprint_test = reproject_to_healpix(
        input_value, "C", nside=64, **kwargs_in
    )

    # Make sure there are some valid values

    assert np.sum(~np.isnan(healpix_data_ref)) == 4

    np.testing.assert_allclose(healpix_data_ref, healpix_data_test)
    np.testing.assert_allclose(footprint_ref, footprint_test)


def test_reproject_from_healpix_output_types(valid_celestial_output_projections):
    wcs_out_ref, shape_ref, output_value, kwargs_out = valid_celestial_output_projections

    array_input = np.random.random(12 * 64**2)

    # Compute reference

    output_ref, footprint_ref = reproject_from_healpix(
        (array_input, "C"), wcs_out_ref, nested=True, shape_out=shape_ref
    )

    # Compute test

    output_test, footprint_test = reproject_from_healpix(
        (array_input, "C"), output_value, nested=True, **kwargs_out
    )

    np.testing.assert_allclose(output_ref, output_test)
    np.testing.assert_allclose(footprint_ref, footprint_test)


def test_reproject_to_healpix_exact_allsky():

    # Regression test for a bug that caused artifacts in the final image if the
    # WCS covered the whole sky - this was due to using scipy's map_coordinates
    # one instead of our built-in one which deals properly with the pixels
    # around the rim.

    shape_out = (160, 320)
    wcs = WCS(naxis=2)
    wcs.wcs.crpix = [(shape_out[1] + 1) / 2, (shape_out[0] + 1) / 2]
    wcs.wcs.cdelt = np.array([-360.0 / shape_out[1], 180.0 / shape_out[0]])
    wcs.wcs.crval = [0, 0]
    wcs.wcs.ctype = ["RA---CAR", "DEC--CAR"]

    array = np.ones(shape_out)

    healpix_array, footprint = reproject_to_healpix(
        (array, wcs),
        coord_system_out="galactic",
        nside=64,
        nested=False,
        order="bilinear",
    )

    assert np.all(footprint > 0)
    assert not np.any(np.isnan(healpix_array))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import datetime
import sys
from importlib import metadata

try:
    from sphinx_astropy.conf.v3 import *  # noqa
except ImportError:
    print("ERROR: the documentation requires the sphinx-astropy package to be installed")
    sys.exit(1)

# -- General configuration ----------------------------------------------------

# By default, highlight as Python 3.
highlight_language = "python3"

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.1'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns.append("_templates")

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog += """
"""

# -- Project information ------------------------------------------------------

package_info = metadata.metadata("reproject")

# This does not *have* to match the package name, but typically does
project = package_info["Name"]
author = package_info["Author"]
copyright = "{}, {}".format(datetime.datetime.now().year, package_info["Author"])

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# The short X.Y version.
version = package_info["Version"].split("-", 1)[0]
# The full version, including alpha/beta/rc tags.
release = package_info["Version"]


# -- Options for HTML output ---------------------------------------------------
# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = f"{project} v{release}"

# Output file base name for HTML help builder.
htmlhelp_basename = project + "doc"


# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [("index", project + ".tex", project + " Documentation", author, "manual")]


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", project.lower(), project + " Documentation", [author], 1)]


## -- Options for the edit_on_github extension ----------------------------------------

nitpicky = True

plot_rcparams = {}
plot_rcparams["figure.figsize"] = (8, 6)
plot_rcparams["savefig.facecolor"] = "none"
plot_rcparams["savefig.bbox"] = "tight"
plot_rcparams["axes.labelsize"] = "large"
plot_rcparams["figure.subplot.hspace"] = 0.5

plot_apply_rcparams = True
plot_html_show_source_link = False
plot_formats = ["png", "svg", "pdf"]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import dask.array as da
import numpy as np
from astropy.wcs import WCS
from astropy.wcs.utils import pixel_to_pixel

from .._array_utils import dask_map_coordinates, map_coordinates
from .._wcs_utils import has_celestial, pixel_to_pixel_with_roundtrip


def _validate_wcs(wcs_in, wcs_out, shape_in, shape_out):
    if wcs_in.low_level_wcs.pixel_n_dim != wcs_out.low_level_wcs.pixel_n_dim:
        raise ValueError("Number of dimensions in input and output WCS should match")
    elif len(shape_out) < wcs_out.low_level_wcs.pixel_n_dim:
        raise ValueError("Too few dimensions in shape_out")
    elif len(shape_in) < wcs_in.low_level_wcs.pixel_n_dim:
        raise ValueError("Too few dimensions in input data")
    elif len(shape_in) != len(shape_out):
        raise ValueError("Number of dimensions in input and output data should match")

    # Separate the "extra" dimensions that don't correspond to a WCS axis and
    # which we'll be looping over
    extra_dimens_in = shape_in[: -wcs_in.low_level_wcs.pixel_n_dim]
    extra_dimens_out = shape_out[: -wcs_out.low_level_wcs.pixel_n_dim]
    if extra_dimens_in != extra_dimens_out:
        raise ValueError("Dimensions to be looped over must match exactly")

    if has_celestial(wcs_in) and not has_celestial(wcs_out):
        raise ValueError("Input WCS has celestial components but output WCS does not")
    elif has_celestial(wcs_out) and not has_celestial(wcs_in):
        raise ValueError("Output WCS has celestial components but input WCS does not")

    if isinstance(wcs_in, WCS) and isinstance(wcs_out, WCS):
        # Check whether a spectral component is present, and if so, check that
        # the CTYPEs match.
        if wcs_in.wcs.spec >= 0 and wcs_out.wcs.spec >= 0:
            if wcs_in.wcs.ctype[wcs_in.wcs.spec][:4] != wcs_out.wcs.ctype[wcs_out.wcs.spec][:4]:
                raise ValueError(
                    f"The input ({wcs_in.wcs.ctype[wcs_in.wcs.spec]}) and output ({wcs_out.wcs.ctype[wcs_out.wcs.spec]}) spectral "
                    "coordinate types are not equivalent."
                )
        elif wcs_in.wcs.spec >= 0:
            raise ValueError("Input WCS has a spectral component but output WCS does not")
        elif wcs_out.wcs.spec >= 0:
            raise ValueError("Output WCS has a spectral component but input WCS does not")


def _reproject_full(
    array,
    wcs_in,
    wcs_out,
    shape_out,
    order=1,
    array_out=None,
    return_footprint=True,
    roundtrip_coords=True,
    output_footprint=None,
):
    """
    Reproject n-dimensional data to a new projection using interpolation.

    The input and output WCS and shape have to satisfy a number of conditions:

    - The number of dimensions in each WCS should match
    - The output shape should match the dimensionality of the WCS
    - The input and output WCS should have matching physical types, although
      the order can be different as long as the physical types are unique.

    If the input array contains extra dimensions beyond what the input WCS has,
    the extra leading dimensions are assumed to represent multiple images with
    the same coordinate information. The transformation is computed once and
    "broadcast" across those images.
    """

    # shape_out must be exactly a tuple type
    shape_out = tuple(shape_out)
    _validate_wcs(wcs_in, wcs_out, array.shape, shape_out)

    is_dask_array = isinstance(array, da.Array)

    if array_out is None:
        array_out = np.empty(shape_out)

    if output_footprint is None:
        output_footprint = np.empty(shape_out)

    array_out_loopable = array_out
    if len(array.shape) == wcs_in.low_level_wcs.pixel_n_dim:
        # We don't need to broadcast the transformation over any extra
        # axes---add an extra axis of length one just so we have something
        # to loop over in all cases.
        array = array.reshape((1, *array.shape))
        array_out_loopable = array_out.reshape((1, *array_out.shape))
    elif len(array.shape) > wcs_in.low_level_wcs.pixel_n_dim:
        # We're broadcasting. Flatten the extra dimensions so there's just one
        # to loop over
        array = array.reshape((-1, *array.shape[-wcs_in.low_level_wcs.pixel_n_dim :]))
        array_out_loopable = array_out.reshape(
            (-1, *array_out.shape[-wcs_out.low_level_wcs.pixel_n_dim :])
        )
    else:
        raise ValueError("Too few dimensions for input array")

    wcs_dims = shape_out[-wcs_in.low_level_wcs.pixel_n_dim :]
    pixel_out = np.meshgrid(
        *[np.arange(size, dtype=float) for size in wcs_dims],
        indexing="ij",
        sparse=False,
        copy=False,
    )
    pixel_out = [p.ravel() for p in pixel_out]
    # For each pixel in the output array, get the pixel value in the input WCS
    if roundtrip_coords:
        pixel_in = pixel_to_pixel_with_roundtrip(wcs_out, wcs_in, *pixel_out[::-1])[::-1]
    else:
        pixel_in = pixel_to_pixel(wcs_out, wcs_in, *pixel_out[::-1])[::-1]
    pixel_in = np.array(pixel_in)

    # Loop over the broadcasted dimensions in our array, reusing the same
    # computed transformation each time
    for i in range(len(array)):
        # Interpolate array on to the pixels coordinates in pixel_in
        if is_dask_array:
            dask_map_coordinates(
                array[i],
                pixel_in,
                order=order,
                cval=np.nan,
                mode="constant",
                output=array_out_loopable[i].ravel(),
            )
        else:
            map_coordinates(
                array[i],
                pixel_in,
                order=order,
                cval=np.nan,
                mode="constant",
                output=array_out_loopable[i].ravel(),
                max_chunk_size=256 * 1024**2,
            )

    # n.b. We write the reprojected data into array_out_loopable, but array_out
    # also contains this data and has the user's desired output shape.

    if return_footprint:
        output_footprint[:] = (~np.isnan(array_out)).astype(float)
        return array_out, output_footprint
    else:
        return array_out



import os
import urllib
from numbers import Number
from pathlib import Path

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord, SpectralCoord
from astropy.wcs.utils import celestial_frame_to_wcs
from astropy_healpix import (
    HEALPix,
    level_to_nside,
)

__all__ = [
    "map_header",
    "tile_header",
    "tile_filename",
    "make_tile_folders",
    "is_url",
    "load_properties",
    "save_properties",
    "spectral_coord_to_index",
    "spectral_index_to_coord",
    "skycoord_first",
]


FREQ_MIN = 1e-18  # Hz
FREQ_MAX = 1e38  # Hz
FREQ_MAX_ORDER = 51


def skycoord_first(worlds):
    """
    Convenience function which takes the output of pixel_to_world and puts
    the SkyCoord first
    """
    for w in worlds:
        if isinstance(w, SkyCoord):
            yield w
    for w in worlds:
        if not isinstance(w, SkyCoord):
            yield w


def spectral_index_to_coord(level, index):
    return SpectralCoord(
        10 ** (index / 2 ** (level + 1) * np.log10(FREQ_MAX / FREQ_MIN) + np.log10(FREQ_MIN)) * u.Hz
    )


def spectral_coord_to_index(level, coord, rounded=True):
    index = (
        2 ** (level + 1) * np.log10(coord.to_value(u.Hz) / FREQ_MIN) / np.log10(FREQ_MAX / FREQ_MIN)
    )
    if rounded:
        return np.floor(index).astype(int)
    else:
        return index


def map_header(*, level, frame, tile_dims):
    if isinstance(level, Number):
        return map_header_2d(
            level=level,
            frame=frame,
            tile_size=tile_dims,
        )
    else:
        return map_header_3d(
            spatial_level=level[0],
            spectral_level=level[1],
            frame=frame,
            tile_size=tile_dims[0],
            tile_depth=tile_dims[1],
        )


def map_header_2d(*, level, frame, tile_size):
    """
    Return the WCS for a whole map stored as a 2D array in HPX projection
    """

    nside = level_to_nside(level)

    # Determine image size
    image_size = 5 * nside * tile_size

    map_wcs = celestial_frame_to_wcs(frame, projection="HPX")
    map_wcs.wcs.crval = 0.0, 0.0

    # Determine map resolution
    res = 45 / tile_size / 2**level
    map_wcs.wcs.cd = [[-res, -res], [res, -res]]

    # Set PV parameters to default values
    map_wcs.wcs.set_pv([(2, 1, 4), (2, 2, 3)])

    # Set origin to center of the image
    map_wcs.wcs.crpix = image_size / 2 + 0.5, image_size / 2 + 0.5

    # Construct header
    header = map_wcs.to_header()

    header["NAXIS"] = 2
    header["WCSAXES"] = 2
    header["NAXIS1"] = image_size
    header["NAXIS2"] = image_size

    return header


def map_header_3d(
    *,
    spatial_level,
    spectral_level,
    frame,
    tile_size,
    tile_depth,
):
    """
    Return the WCS for a whole map stored as a 3D array in HPX projection
    """

    # First get the 2D header
    header = map_header_2d(level=spatial_level, frame=frame, tile_size=tile_size)

    # Then modify it to be 3D
    header["NAXIS"] = 3
    header["WCSAXES"] = 3
    header["NAXIS3"] = tile_depth * 2 ** (spectral_level + 1)
    header["FORDER"] = spectral_level
    header["CTYPE3"] = "FREQ-LOG"
    header["CUNIT3"] = "Hz"
    header["CRPIX3"] = 1
    header["CRVAL3"] = FREQ_MIN
    header["CDELT3"] = (
        FREQ_MIN * np.log(FREQ_MAX / FREQ_MIN) / 2 ** (spectral_level + 1 + np.log2(tile_depth))
    )

    return header


def tile_header(*, level, index, frame, tile_dims):
    if isinstance(level, Number):
        return tile_header_2d(
            level=level,
            index=index,
            frame=frame,
            tile_size=tile_dims,
        )
    else:
        return tile_header_3d(
            spatial_level=level[0],
            spatial_index=index[0],
            spectral_level=level[1],
            spectral_index=index[1],
            frame=frame,
            tile_size=tile_dims[0],
            tile_depth=tile_dims[1],
        )


def tile_header_2d(*, level, index, frame, tile_size):
    """
    Return the WCS for a given HiPS tile
    """

    # PERF: we could optimize this by making it into a class, as very little
    # changes from tile to tile so we could first create the base tile header
    # and then just update values for each tile for performance.

    nside = level_to_nside(level)

    hp = HEALPix(nside=nside, order="nested", frame=frame)

    tile_wcs = celestial_frame_to_wcs(frame, projection="HPX")
    tile_wcs.wcs.crval = 0.0, 0.0

    # Determine tile resolution
    res = 45 / tile_size / 2**level  # degrees
    tile_wcs.wcs.cd = [[-res, -res], [res, -res]]

    # Set PV parameters to default values
    tile_wcs.wcs.set_pv([(2, 1, 4), (2, 2, 3)])

    # Determine CRPIX values by determining the position of the relevant corner
    # relative to the origin of the projection.
    offset_x, offset_y = tile_wcs.world_to_pixel(
        hp.healpix_to_skycoord(index, dx=[0.5, 0.9, 0.9, 0.1, 0.1], dy=[0.5, 0.1, 0.9, 0.9, 0.1])
    )
    border_tile = (
        np.max(np.hypot(offset_x[1:] - offset_x[0], offset_y[1:] - offset_y[0])) > tile_size
    )

    offset_x, offset_y = tile_wcs.world_to_pixel(hp.healpix_to_skycoord(index, dx=0.75, dy=0.25))

    tile_wcs.wcs.crpix[0] = -offset_x - 0.5 + tile_size / 4
    tile_wcs.wcs.crpix[1] = -offset_y - 0.5 + tile_size / 4

    # Construct header
    header = tile_wcs.to_header()

    header["NPIX"] = index
    header["ORDER"] = level
    header["NAXIS"] = 2
    header["WCSAXES"] = 2
    header["NAXIS1"] = tile_size
    header["NAXIS2"] = tile_size

    if border_tile:
        header2 = header.copy()
        header2["CRPIX1"] = -header["CRPIX2"] + tile_size + 1
        header2["CRPIX2"] = -header["CRPIX1"] + tile_size + 1
        return header, header2
    else:
        return header


def tile_header_3d(
    *,
    spatial_level,
    spatial_index,
    spectral_level,
    spectral_index,
    frame,
    tile_size,
    tile_depth,
):

    # First get the 2D header
    headers = tile_header_2d(
        level=spatial_level, index=spatial_index, frame=frame, tile_size=tile_size
    )

    if not isinstance(headers, tuple):
        headers = (headers,)

    for header in headers:

        # Then modify it to be 3D
        header["NAXIS"] = 3
        header["WCSAXES"] = 3
        header["NAXIS3"] = tile_depth
        header["FORDER"] = spectral_level
        header["FPIX"] = spectral_index
        header["CTYPE3"] = "FREQ-LOG"
        header["CUNIT3"] = "Hz"
        header["CRPIX3"] = 1 - spectral_index * tile_depth
        header["CRVAL3"] = FREQ_MIN
        header["CDELT3"] = (
            FREQ_MIN * np.log(FREQ_MAX / FREQ_MIN) / 2 ** (spectral_level + 1 + np.log2(tile_depth))
        )

    if len(headers) == 1:
        return headers[0]
    else:
        return headers


def _rounded_spatial_index(index):
    return 10000 * (index // 10000)


def _rounded_spectral_index(index):
    return 10 * (index // 10)


def tile_filename(*, level, index, output_directory, extension):
    if isinstance(level, Number):
        return tile_filename_2d(
            level=level, index=index, output_directory=output_directory, extension=extension
        )
    else:
        return tile_filename_3d(
            spatial_level=level[0],
            spatial_index=index[0],
            spectral_level=level[1],
            spectral_index=index[1],
            output_directory=output_directory,
            extension=extension,
        )


def tile_filename_2d(*, level, index, output_directory, extension):
    return os.path.join(
        output_directory,
        f"Norder{level}",
        f"Dir{_rounded_spatial_index(index)}",
        f"Npix{index}.{extension}",
    )


def tile_filename_3d(
    *, spatial_level, spectral_level, spatial_index, spectral_index, output_directory, extension
):
    return os.path.join(
        output_directory,
        f"Norder{spatial_level}_{spectral_level}",
        f"Dir{_rounded_spatial_index(spatial_index)}_{_rounded_spectral_index(spectral_index)}",
        f"Npix{spatial_index}_{spectral_index}.{extension}",
    )


def make_tile_folders(*, level, indices, output_directory):
    if isinstance(level, Number):
        make_tile_folders_2d(
            level=level,
            indices=indices,
            output_directory=output_directory,
        )
    else:
        indices = list(zip(*indices, strict=False))
        make_tile_folders_3d(
            spatial_level=level[0],
            spectral_level=level[1],
            spatial_indices=np.array(indices[0]),
            spectral_indices=np.array(indices[1]),
            output_directory=output_directory,
        )


def make_tile_folders_2d(*, level, indices, output_directory):

    rounded_indices = np.unique(_rounded_spatial_index(indices))
    for index in rounded_indices:
        dirname = os.path.dirname(
            tile_filename(level=level, index=index, output_directory=output_directory, extension="")
        )
        if not os.path.exists(dirname):
            os.makedirs(dirname)


def is_url(directory):
    if isinstance(directory, Path):
        return False
    else:
        return directory.startswith("http://") or directory.startswith("https://")


def make_tile_folders_3d(
    *, spatial_level, spectral_level, spatial_indices, spectral_indices, output_directory
):
    rounded_spatial_indices = np.unique(_rounded_spatial_index(spatial_indices))
    for spatial_index in rounded_spatial_indices:
        for spectral_index in np.unique(_rounded_spectral_index(spectral_indices)):
            dirname = os.path.dirname(
                tile_filename_3d(
                    spatial_level=spatial_level,
                    spectral_level=spectral_level,
                    spatial_index=spatial_index,
                    spectral_index=spectral_index,
                    output_directory=output_directory,
                    extension="",
                )
            )
            if not os.path.exists(dirname):
                os.makedirs(dirname)


def save_properties(directory, properties):
    with open(os.path.join(directory, "properties"), "w") as f:
        for key, value in properties.items():
            f.write(f"{key:20s} = {value}\n")


def load_properties(directory_or_url):

    if is_url(directory_or_url):
        properties_filename, _ = urllib.request.urlretrieve(f"{directory_or_url}/properties")
    else:
        properties_filename = os.path.join(directory_or_url, "properties")

    properties = {}
    with open(properties_filename) as f:
        for line in f:
            if line.startswith("#") or line.strip() == "":
                continue
            key, value = line.split("=", 1)
            properties[key.strip()] = value.strip()

    return properties



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import itertools
import logging

import dask.array as da
import numpy as np
import pytest
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS
from astropy.wcs.wcs import FITSFixedWarning
from astropy.wcs.wcsapi import HighLevelWCSWrapper, SlicedLowLevelWCS
from numpy.testing import assert_allclose

from reproject._array_utils import ArrayWrapper
from reproject.interpolation._high_level import reproject_interp
from reproject.tests.helpers import array_footprint_to_hdulist

# TODO: add reference comparisons


@pytest.fixture(
    params=[None, "memmap", "none"],
)
def dask_method(request):
    return request.param


def as_high_level_wcs(wcs):
    return HighLevelWCSWrapper(SlicedLowLevelWCS(wcs, Ellipsis))


@pytest.mark.array_compare(single_reference=True)
@pytest.mark.parametrize("wcsapi", (False, True))
@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_reproject_celestial_2d_gal2equ(wcsapi, roundtrip_coords):
    """
    Test reprojection of a 2D celestial image, which includes a coordinate
    system conversion.
    """
    with fits.open(get_pkg_data_filename("data/galactic_2d.fits", package="reproject.tests")) as pf:
        hdu_in = pf[0]
        header_out = hdu_in.header.copy()
        header_out["CTYPE1"] = "RA---TAN"
        header_out["CTYPE2"] = "DEC--TAN"
        header_out["CRVAL1"] = 266.39311
        header_out["CRVAL2"] = -28.939779

        if wcsapi:  # Enforce a pure wcsapi API
            wcs_in, data_in = as_high_level_wcs(WCS(hdu_in.header)), hdu_in.data
            wcs_out = as_high_level_wcs(WCS(header_out))
            shape_out = header_out["NAXIS2"], header_out["NAXIS1"]
            array_out, footprint_out = reproject_interp(
                (data_in, wcs_in), wcs_out, shape_out=shape_out, roundtrip_coords=roundtrip_coords
            )
        else:
            array_out, footprint_out = reproject_interp(
                hdu_in, header_out, roundtrip_coords=roundtrip_coords
            )

    return array_footprint_to_hdulist(array_out, footprint_out, header_out)


# Note that we can't use independent_celestial_slices=True and reorder the
# axes, hence why we need to prepare the combinations in this way.
AXIS_ORDER = list(itertools.permutations((0, 1, 2)))
COMBINATIONS = []
for wcsapi in (False, True):
    for axis_order in AXIS_ORDER:
        COMBINATIONS.append((wcsapi, axis_order))


@pytest.mark.array_compare(single_reference=True)
@pytest.mark.parametrize(("wcsapi", "axis_order"), tuple(COMBINATIONS))
@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_reproject_celestial_3d_equ2gal(wcsapi, axis_order, roundtrip_coords):
    """
    Test reprojection of a 3D cube with celestial components, which includes a
    coordinate system conversion (the original header is in equatorial
    coordinates). We test using both the 'fast' method which assumes celestial
    slices are independent, and the 'full' method. We also scramble the input
    dimensions of the data and header to make sure that the reprojection can
    deal with this.
    """

    # Read in the input cube
    with fits.open(
        get_pkg_data_filename("data/equatorial_3d.fits", package="reproject.tests")
    ) as pf:
        hdu_in = pf[0]

        # Define the output header - this should be the same for all versions of
        # this test to make sure we can use a single reference file.
        header_out = hdu_in.header.copy()
        header_out["NAXIS1"] = 10
        header_out["NAXIS2"] = 9
        header_out["CTYPE1"] = "GLON-SIN"
        header_out["CTYPE2"] = "GLAT-SIN"
        header_out["CRVAL1"] = 163.16724
        header_out["CRVAL2"] = -15.777405
        header_out["CRPIX1"] = 6
        header_out["CRPIX2"] = 5

        # We now scramble the input axes
        if axis_order != (0, 1, 2):
            wcs_in = WCS(hdu_in.header)
            wcs_in = wcs_in.sub((3 - np.array(axis_order)[::-1]).tolist())
            hdu_in.header = wcs_in.to_header()
            hdu_in.data = np.transpose(hdu_in.data, axis_order)

        if wcsapi:  # Enforce a pure wcsapi API
            wcs_in, data_in = as_high_level_wcs(WCS(hdu_in.header)), hdu_in.data
            wcs_out = as_high_level_wcs(WCS(header_out))
            shape_out = header_out["NAXIS3"], header_out["NAXIS2"], header_out["NAXIS1"]
            array_out, footprint_out = reproject_interp(
                (data_in, wcs_in), wcs_out, shape_out=shape_out, roundtrip_coords=roundtrip_coords
            )
        else:
            array_out, footprint_out = reproject_interp(
                hdu_in, header_out, roundtrip_coords=roundtrip_coords
            )

    return array_footprint_to_hdulist(array_out, footprint_out, header_out)


@pytest.mark.array_compare(single_reference=True)
@pytest.mark.parametrize("wcsapi", (False, True))
@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_small_cutout(wcsapi, roundtrip_coords):
    """
    Test reprojection of a cutout from a larger image (makes sure that the
    pre-reprojection cropping works)
    """
    with fits.open(get_pkg_data_filename("data/galactic_2d.fits", package="reproject.tests")) as pf:
        hdu_in = pf[0]
        header_out = hdu_in.header.copy()
        header_out["NAXIS1"] = 10
        header_out["NAXIS2"] = 9
        header_out["CTYPE1"] = "RA---TAN"
        header_out["CTYPE2"] = "DEC--TAN"
        header_out["CRVAL1"] = 266.39311
        header_out["CRVAL2"] = -28.939779
        header_out["CRPIX1"] = 5.1
        header_out["CRPIX2"] = 4.7

        if wcsapi:  # Enforce a pure wcsapi API
            wcs_in, data_in = as_high_level_wcs(WCS(hdu_in.header)), hdu_in.data
            wcs_out = as_high_level_wcs(WCS(header_out))
            shape_out = header_out["NAXIS2"], header_out["NAXIS1"]
            array_out, footprint_out = reproject_interp(
                (data_in, wcs_in), wcs_out, shape_out=shape_out, roundtrip_coords=roundtrip_coords
            )
        else:
            array_out, footprint_out = reproject_interp(
                hdu_in, header_out, roundtrip_coords=roundtrip_coords
            )

    return array_footprint_to_hdulist(array_out, footprint_out, header_out)


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_mwpan_car_to_mol(roundtrip_coords):
    """
    Test reprojection of the Mellinger Milky Way Panorama from CAR to MOL,
    which was returning all NaNs due to a regression that was introduced in
    reproject 0.3 (https://github.com/astrofrog/reproject/pull/124).
    """
    hdu_in = fits.Header.fromtextfile(
        get_pkg_data_filename("data/mwpan2_RGB_3600.hdr", package="reproject.tests")
    )
    with pytest.warns(FITSFixedWarning):
        wcs_in = WCS(hdu_in, naxis=2)
    data_in = np.ones((hdu_in["NAXIS2"], hdu_in["NAXIS1"]), dtype=float)
    header_out = fits.Header()
    header_out["NAXIS"] = 2
    header_out["NAXIS1"] = 360
    header_out["NAXIS2"] = 180
    header_out["CRPIX1"] = 180
    header_out["CRPIX2"] = 90
    header_out["CRVAL1"] = 0
    header_out["CRVAL2"] = 0
    header_out["CDELT1"] = -2 * np.sqrt(2) / np.pi
    header_out["CDELT2"] = 2 * np.sqrt(2) / np.pi
    header_out["CTYPE1"] = "GLON-MOL"
    header_out["CTYPE2"] = "GLAT-MOL"
    header_out["RADESYS"] = "ICRS"
    array_out, footprint_out = reproject_interp(
        (data_in, wcs_in), header_out, roundtrip_coords=roundtrip_coords
    )
    assert np.isfinite(array_out).any()


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_small_cutout_outside(roundtrip_coords):
    """
    Test reprojection of a cutout from a larger image - in this case the
    cutout is completely outside the region of the input image so we should
    take a shortcut that returns arrays of NaNs.
    """
    with fits.open(get_pkg_data_filename("data/galactic_2d.fits", package="reproject.tests")) as pf:
        hdu_in = pf[0]
        header_out = hdu_in.header.copy()
        header_out["NAXIS1"] = 10
        header_out["NAXIS2"] = 9
        header_out["CTYPE1"] = "RA---TAN"
        header_out["CTYPE2"] = "DEC--TAN"
        header_out["CRVAL1"] = 216.39311
        header_out["CRVAL2"] = -21.939779
        header_out["CRPIX1"] = 5.1
        header_out["CRPIX2"] = 4.7
        array_out, footprint_out = reproject_interp(
            hdu_in, header_out, roundtrip_coords=roundtrip_coords
        )
    assert np.all(np.isnan(array_out))
    assert np.all(footprint_out == 0)


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_celestial_mismatch_2d(roundtrip_coords):
    """
    Make sure an error is raised if the input image has celestial WCS
    information and the output does not (and vice-versa). This example will
    use the _reproject_celestial route.
    """
    with fits.open(get_pkg_data_filename("data/galactic_2d.fits", package="reproject.tests")) as pf:
        hdu_in = pf[0]

        header_out = hdu_in.header.copy()
        header_out["CTYPE1"] = "APPLES"
        header_out["CTYPE2"] = "ORANGES"

        data = hdu_in.data
        wcs1 = WCS(hdu_in.header)
        wcs2 = WCS(header_out)

        with pytest.raises(
            ValueError, match="Input WCS has celestial components but output WCS does not"
        ):
            array_out, footprint_out = reproject_interp(
                (data, wcs1), wcs2, shape_out=(2, 2), roundtrip_coords=roundtrip_coords
            )


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_celestial_mismatch_3d(roundtrip_coords):
    """
    Make sure an error is raised if the input image has celestial WCS
    information and the output does not (and vice-versa). This example will
    use the _reproject_full route.
    """
    with fits.open(
        get_pkg_data_filename("data/equatorial_3d.fits", package="reproject.tests")
    ) as pf:
        hdu_in = pf[0]

        header_out = hdu_in.header.copy()
        header_out["CTYPE1"] = "APPLES"
        header_out["CTYPE2"] = "ORANGES"
        header_out["CTYPE3"] = "BANANAS"

        data = hdu_in.data
        wcs1 = WCS(hdu_in.header)
        wcs2 = WCS(header_out)

        with pytest.raises(
            ValueError, match="Input WCS has celestial components but output WCS does not"
        ):
            array_out, footprint_out = reproject_interp(
                (data, wcs1), wcs2, shape_out=(1, 2, 3), roundtrip_coords=roundtrip_coords
            )

        with pytest.raises(
            ValueError, match="Output WCS has celestial components but input WCS does not"
        ):
            array_out, footprint_out = reproject_interp(
                (data, wcs2), wcs1, shape_out=(1, 2, 3), roundtrip_coords=roundtrip_coords
            )


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_spectral_mismatch_3d(roundtrip_coords):
    """
    Make sure an error is raised if there are mismatches between the presence
    or type of spectral axis.
    """
    with fits.open(
        get_pkg_data_filename("data/equatorial_3d.fits", package="reproject.tests")
    ) as pf:
        hdu_in = pf[0]

        header_out = hdu_in.header.copy()
        header_out["CTYPE3"] = "FREQ"
        header_out["CUNIT3"] = "Hz"

        data = hdu_in.data
        wcs1 = WCS(hdu_in.header)
        wcs2 = WCS(header_out)

        with pytest.raises(
            ValueError,
            match=r"The input \(VOPT\) and output \(FREQ\) spectral "
            r"coordinate types are not equivalent\.",
        ):
            array_out, footprint_out = reproject_interp(
                (data, wcs1), wcs2, shape_out=(1, 2, 3), roundtrip_coords=roundtrip_coords
            )

        header_out["CTYPE3"] = "BANANAS"
        wcs2 = WCS(header_out)

        with pytest.raises(
            ValueError, match="Input WCS has a spectral component but output WCS does not"
        ):
            array_out, footprint_out = reproject_interp(
                (data, wcs1), wcs2, shape_out=(1, 2, 3), roundtrip_coords=roundtrip_coords
            )

        with pytest.raises(
            ValueError, match="Output WCS has a spectral component but input WCS does not"
        ):
            array_out, footprint_out = reproject_interp(
                (data, wcs2), wcs1, shape_out=(1, 2, 3), roundtrip_coords=roundtrip_coords
            )


@pytest.mark.parametrize("roundtrip_coords", (False, True))
def test_naxis_mismatch(roundtrip_coords):
    """
    Make sure an error is raised if the input and output WCS have a different
    number of dimensions.
    """
    data = np.ones((3, 2, 2))
    wcs_in = WCS(naxis=3)
    wcs_out = WCS(naxis=2)

    with pytest.raises(
        ValueError, match="Number of dimensions in input and output WCS should match"
    ):
        array_out, footprint_out = reproject_interp(
            (data, wcs_in), wcs_out, shape_out=(1, 2), roundtrip_coords=roundtrip_coords
        )


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_slice_reprojection(roundtrip_coords):
    """
    Test case where only the slices change and the celestial projection doesn't
    """
    inp_cube = np.arange(3, dtype="float").repeat(4 * 5).reshape(3, 4, 5)

    header_in = fits.Header.fromtextfile(
        get_pkg_data_filename("data/cube.hdr", package="reproject.tests")
    )

    header_in["NAXIS1"] = 5
    header_in["NAXIS2"] = 4
    header_in["NAXIS3"] = 3

    header_out = header_in.copy()
    header_out["NAXIS3"] = 2
    header_out["CRPIX3"] -= 0.5

    wcs_in = WCS(header_in)
    wcs_out = WCS(header_out)

    out_cube, out_cube_valid = reproject_interp(
        (inp_cube, wcs_in), wcs_out, shape_out=(2, 4, 5), roundtrip_coords=roundtrip_coords
    )

    # we expect to be projecting from
    # inp_cube = np.arange(3, dtype='float').repeat(4*5).reshape(3,4,5)
    # to
    # inp_cube_interp = (inp_cube[:-1]+inp_cube[1:])/2.
    # which is confirmed by
    # map_coordinates(inp_cube.astype('float'), new_coords, order=1, cval=np.nan, mode='constant')
    # np.testing.assert_allclose(inp_cube_interp, map_coordinates(inp_cube.astype('float'),
    # new_coords, order=1, cval=np.nan, mode='constant'))

    assert out_cube.shape == (2, 4, 5)
    assert out_cube_valid.sum() == 40.0

    # We only check that the *valid* pixels are equal
    # but it's still nice to check that the "valid" array works as a mask
    np.testing.assert_allclose(
        out_cube[out_cube_valid.astype("bool")],
        ((inp_cube[:-1] + inp_cube[1:]) / 2.0)[out_cube_valid.astype("bool")],
    )

    # Actually, I fixed it, so now we can test all
    np.testing.assert_allclose(out_cube, ((inp_cube[:-1] + inp_cube[1:]) / 2.0))


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_inequal_wcs_dims(roundtrip_coords):
    inp_cube = np.arange(3, dtype="float").repeat(4 * 5).reshape(3, 4, 5)
    header_in = fits.Header.fromtextfile(
        get_pkg_data_filename("data/cube.hdr", package="reproject.tests")
    )

    header_out = header_in.copy()
    header_out["CTYPE3"] = "VRAD"
    header_out["CUNIT3"] = "m/s"
    header_in["CTYPE3"] = "STOKES"
    header_in["CUNIT3"] = ""

    wcs_out = WCS(header_out)

    with pytest.raises(
        ValueError, match="Output WCS has a spectral component but input WCS does not"
    ):
        out_cube, out_cube_valid = reproject_interp(
            (inp_cube, header_in), wcs_out, shape_out=(2, 4, 5), roundtrip_coords=roundtrip_coords
        )


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_different_wcs_types(roundtrip_coords):
    inp_cube = np.arange(3, dtype="float").repeat(4 * 5).reshape(3, 4, 5)
    header_in = fits.Header.fromtextfile(
        get_pkg_data_filename("data/cube.hdr", package="reproject.tests")
    )

    header_out = header_in.copy()
    header_out["CTYPE3"] = "VRAD"
    header_out["CUNIT3"] = "m/s"
    header_in["CTYPE3"] = "VELO"
    header_in["CUNIT3"] = "m/s"

    wcs_out = WCS(header_out)

    with pytest.raises(
        ValueError,
        match=r"The input \(VELO\) and output \(VRAD\) spectral "
        r"coordinate types are not equivalent\.",
    ):
        out_cube, out_cube_valid = reproject_interp(
            (inp_cube, header_in), wcs_out, shape_out=(2, 4, 5), roundtrip_coords=roundtrip_coords
        )


# TODO: add a test to check the units are the same.


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_reproject_3d_celestial_correctness_ra2gal(roundtrip_coords):
    inp_cube = np.arange(3, dtype="float").repeat(7 * 8).reshape(3, 7, 8)

    header_in = fits.Header.fromtextfile(
        get_pkg_data_filename("data/cube.hdr", package="reproject.tests")
    )

    header_in["NAXIS1"] = 8
    header_in["NAXIS2"] = 7
    header_in["NAXIS3"] = 3

    header_out = header_in.copy()
    header_out["CTYPE1"] = "GLON-TAN"
    header_out["CTYPE2"] = "GLAT-TAN"
    header_out["CRVAL1"] = 158.5644791
    header_out["CRVAL2"] = -21.59589875
    # make the cube a cutout approximately in the center of the other one, but smaller
    header_out["NAXIS1"] = 4
    header_out["CRPIX1"] = 2
    header_out["NAXIS2"] = 3
    header_out["CRPIX2"] = 1.5

    header_out["NAXIS3"] = 2
    header_out["CRPIX3"] -= 0.5

    wcs_in = WCS(header_in)
    wcs_out = WCS(header_out)

    out_cube, out_cube_valid = reproject_interp(
        (inp_cube, wcs_in), wcs_out, shape_out=(2, 3, 4), roundtrip_coords=roundtrip_coords
    )

    assert out_cube.shape == (2, 3, 4)
    assert out_cube_valid.sum() == out_cube.size

    # only compare the spectral axis
    np.testing.assert_allclose(out_cube[:, 0, 0], ((inp_cube[:-1] + inp_cube[1:]) / 2.0)[:, 0, 0])


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_reproject_with_output_array(roundtrip_coords):
    """
    Test both full_reproject and slicewise reprojection. We use a case where the
    non-celestial slices are the same and therefore where both algorithms can
    work.
    """
    header_in = fits.Header.fromtextfile(
        get_pkg_data_filename("data/cube.hdr", package="reproject.tests")
    )

    array_in = np.ones((3, 200, 180))
    shape_out = (3, 160, 170)
    out_full = np.empty(shape_out)

    wcs_in = WCS(header_in)
    wcs_out = wcs_in.deepcopy()
    wcs_out.wcs.ctype = ["GLON-SIN", "GLAT-SIN", wcs_in.wcs.ctype[2]]
    wcs_out.wcs.crval = [158.0501, -21.530282, wcs_in.wcs.crval[2]]
    wcs_out.wcs.crpix = [50.0, 50.0, wcs_in.wcs.crpix[2] + 0.4]

    # TODO when someone learns how to do it: make sure the memory isn't duplicated...
    returned_array = reproject_interp(
        (array_in, wcs_in),
        wcs_out,
        output_array=out_full,
        return_footprint=False,
        roundtrip_coords=roundtrip_coords,
    )

    assert out_full is returned_array


@pytest.mark.array_compare(single_reference=True)
@pytest.mark.remote_data
def test_reproject_roundtrip(aia_test_data):
    # Test the reprojection with solar data, which ensures that the masking of
    # pixels based on round-tripping works correctly. Using asdf is not just
    # about testing a different format but making sure that GWCS works.

    pytest.importorskip("sunpy", minversion="6.0.1")

    data, wcs, target_wcs = aia_test_data

    output, footprint = reproject_interp((data, wcs), target_wcs, (128, 128))

    header_out = target_wcs.to_header()

    header_out["DATE-OBS"] = header_out["DATE-OBS"].replace("T", " ")

    # With sunpy 6.0.0 and later, additional keyword arguments are written out
    # so we remove these as they are not important for the comparison with the
    # reference files.
    header_out.pop("DATE-AVG", None)
    header_out.pop("MJD-AVG", None)

    return array_footprint_to_hdulist(output, footprint, header_out)


def test_reproject_roundtrip_kwarg(aia_test_data):
    # Make sure that the roundtrip_coords keyword argument has an effect. This
    # is a regression test for a bug that caused the keyword argument to be
    # ignored when in parallel/blocked mode.

    pytest.importorskip("sunpy", minversion="6.0.1")

    data, wcs, target_wcs = aia_test_data

    output_roundtrip_1 = reproject_interp(
        (data, wcs), target_wcs, shape_out=(128, 128), return_footprint=False, roundtrip_coords=True
    )
    output_roundtrip_2 = reproject_interp(
        (data, wcs),
        target_wcs,
        shape_out=(128, 128),
        return_footprint=False,
        roundtrip_coords=True,
        block_size=(32, 32),
    )

    assert_allclose(output_roundtrip_1, output_roundtrip_2)

    output_noroundtrip_1 = reproject_interp(
        (data, wcs),
        target_wcs,
        shape_out=(128, 128),
        return_footprint=False,
        roundtrip_coords=False,
    )
    output_noroundtrip_2 = reproject_interp(
        (data, wcs),
        target_wcs,
        shape_out=(128, 128),
        return_footprint=False,
        roundtrip_coords=False,
        block_size=(32, 32),
    )

    assert_allclose(output_noroundtrip_1, output_noroundtrip_2)

    # The array with round-tripping should have more NaN values:
    assert np.sum(np.isnan(output_roundtrip_1)) > 9500
    assert np.sum(np.isnan(output_noroundtrip_1)) < 7000


@pytest.mark.parametrize("roundtrip_coords", (False, True))
@pytest.mark.remote_data
def test_identity_with_offset(roundtrip_coords):
    # Reproject an array and WCS to itself but with a margin, which should
    # end up empty. This is a regression test for a bug that caused some
    # values to extend beyond the original footprint.

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "RA---TAN", "DEC--TAN"
    wcs.wcs.crpix = 322, 151
    wcs.wcs.crval = 43, 23
    wcs.wcs.cdelt = -0.1, 0.1
    wcs.wcs.equinox = 2000.0

    array_in = np.random.random((233, 123))

    wcs_out = wcs.deepcopy()
    wcs_out.wcs.crpix += 1
    shape_out = (array_in.shape[0] + 2, array_in.shape[1] + 2)

    array_out, footprint = reproject_interp(
        (array_in, wcs), wcs_out, shape_out=shape_out, roundtrip_coords=roundtrip_coords
    )

    expected = np.pad(array_in, 1, "constant", constant_values=np.nan)

    assert_allclose(expected, array_out, atol=1e-10)


def _setup_for_broadcast_test():
    with fits.open(get_pkg_data_filename("data/galactic_2d.fits", package="reproject.tests")) as pf:
        hdu_in = pf[0]
        header_in = hdu_in.header.copy()
        header_out = hdu_in.header.copy()
        header_out["CTYPE1"] = "RA---TAN"
        header_out["CTYPE2"] = "DEC--TAN"
        header_out["CRVAL1"] = 266.39311
        header_out["CRVAL2"] = -28.939779

        data = hdu_in.data

    image_stack = np.stack((data, data.T, data[::-1], data[:, ::-1]))

    # Build the reference array through un-broadcast reprojections
    array_ref = np.empty_like(image_stack)
    footprint_ref = np.empty_like(image_stack)
    for i in range(len(image_stack)):
        array_out, footprint_out = reproject_interp((image_stack[i], header_in), header_out)
        array_ref[i] = array_out
        footprint_ref[i] = footprint_out

    return image_stack, array_ref, footprint_ref, header_in, header_out


@pytest.mark.parametrize("input_extra_dims", (1, 2))
@pytest.mark.parametrize("output_shape", (None, "single", "full"))
@pytest.mark.parametrize("input_as_wcs", (True, False))
@pytest.mark.parametrize("output_as_wcs", (True, False))
def test_broadcast_reprojection(input_extra_dims, output_shape, input_as_wcs, output_as_wcs):
    image_stack, array_ref, footprint_ref, header_in, header_out = _setup_for_broadcast_test()
    # Test both single and multiple dimensions being broadcast
    if input_extra_dims == 2:
        image_stack = image_stack.reshape((2, 2, *image_stack.shape[-2:]))
        array_ref = array_ref.reshape(image_stack.shape)
        footprint_ref = footprint_ref.reshape(image_stack.shape)

    # Test different ways of providing the output shape
    if output_shape == "single":
        # Have the broadcast dimensions be auto-added to the output shape
        output_shape = image_stack.shape[-2:]
    elif output_shape == "full":
        # Provide the broadcast dimensions as part of the output shape
        output_shape = image_stack.shape

    # Ensure logic works with WCS inputs as well as Header inputs
    if input_as_wcs:
        header_in = WCS(header_in)
    if output_as_wcs:
        header_out = WCS(header_out)
        if output_shape is None:
            # This combination of parameter values is not valid
            return

    array_broadcast, footprint_broadcast = reproject_interp(
        (image_stack, header_in),
        header_out,
        output_shape,
    )

    np.testing.assert_array_equal(footprint_broadcast, footprint_ref)
    np.testing.assert_allclose(array_broadcast, array_ref)


# In the tests below we ignore FITSFixedWarning due to:
# https://github.com/astropy/astropy/pull/12844


@pytest.mark.parametrize("input_extra_dims", (1, 2))
@pytest.mark.parametrize("output_shape", (None, "single", "full"))
@pytest.mark.parametrize("parallel", [True, False])
@pytest.mark.parametrize("header_or_wcs", (lambda x: x, WCS))
@pytest.mark.filterwarnings("ignore::astropy.wcs.wcs.FITSFixedWarning")
def test_blocked_broadcast_reprojection(input_extra_dims, output_shape, parallel, header_or_wcs):
    image_stack, array_ref, footprint_ref, header_in, header_out = _setup_for_broadcast_test()
    # Test both single and multiple dimensions being broadcast
    if input_extra_dims == 2:
        image_stack = image_stack.reshape((2, 2, *image_stack.shape[-2:]))
        array_ref = array_ref.reshape(image_stack.shape)
        footprint_ref = footprint_ref.reshape(image_stack.shape)

    # Test different ways of providing the output shape
    if output_shape == "single":
        # Have the broadcast dimensions be auto-added to the output shape
        output_shape = image_stack.shape[-2:]
    elif output_shape == "full":
        # Provide the broadcast dimensions as part of the output shape
        output_shape = image_stack.shape

    # test different behavior when the output projection is a WCS
    header_out = header_or_wcs(header_out)

    array_broadcast, footprint_broadcast = reproject_interp(
        (image_stack, header_in), header_out, output_shape, parallel=parallel, block_size=[5, 5]
    )

    np.testing.assert_array_equal(footprint_broadcast, footprint_ref)
    np.testing.assert_allclose(array_broadcast, array_ref)


@pytest.mark.parametrize("parallel", [True, 2, False])
@pytest.mark.parametrize("block_size", [[500, 500], [500, 100], None])
@pytest.mark.parametrize("return_footprint", [False, True])
@pytest.mark.parametrize("existing_outputs", [False, True])
@pytest.mark.parametrize("header_or_wcs", (lambda x: x, WCS))
@pytest.mark.remote_data
@pytest.mark.filterwarnings("ignore::astropy.wcs.wcs.FITSFixedWarning")
def test_blocked_against_single(
    parallel, block_size, return_footprint, existing_outputs, header_or_wcs
):
    # Ensure when we break a reprojection down into multiple discrete blocks
    # it has the same result as if all pixels where reprejcted at once

    hdu1 = fits.open(get_pkg_data_filename("galactic_center/gc_2mass_k.fits"))[0]
    hdu2 = fits.open(get_pkg_data_filename("galactic_center/gc_msx_e.fits"))[0]
    array_test = None
    footprint_test = None

    shape_out = (720, 721)

    if existing_outputs:
        output_array_test = np.zeros(shape_out)
        output_footprint_test = np.zeros(shape_out)
        output_array_reference = np.zeros(shape_out)
        output_footprint_reference = np.zeros(shape_out)
    else:
        output_array_test = None
        output_footprint_test = None
        output_array_reference = None
        output_footprint_reference = None

    result_test = reproject_interp(
        hdu2,
        header_or_wcs(hdu1.header),
        parallel=parallel,
        block_size=block_size,
        return_footprint=return_footprint,
        output_array=output_array_test,
        output_footprint=output_footprint_test,
    )

    result_reference = reproject_interp(
        hdu2,
        header_or_wcs(hdu1.header),
        parallel=False,
        block_size=None,
        return_footprint=return_footprint,
        output_array=output_array_reference,
        output_footprint=output_footprint_reference,
    )

    if return_footprint:
        array_test, footprint_test = result_test
        array_reference, footprint_reference = result_reference
    else:
        array_test = result_test
        array_reference = result_reference

    if existing_outputs:
        assert array_test is output_array_test
        assert array_reference is output_array_reference
        if return_footprint:
            assert footprint_test is output_footprint_test
            assert footprint_reference is output_footprint_reference

    np.testing.assert_allclose(array_test, array_reference, equal_nan=True)
    if return_footprint:
        np.testing.assert_allclose(footprint_test, footprint_reference, equal_nan=True)


def test_interp_input_output_types(valid_celestial_input_data, valid_celestial_output_projections):
    # Check that all valid input/output types work properly

    array_ref, wcs_in_ref, input_value, kwargs_in = valid_celestial_input_data

    wcs_out_ref, shape_ref, output_value, kwargs_out = valid_celestial_output_projections

    # Compute reference

    output_ref, footprint_ref = reproject_interp(
        (array_ref, wcs_in_ref), wcs_out_ref, shape_out=shape_ref
    )

    # Compute test

    output_test, footprint_test = reproject_interp(
        input_value, output_value, **kwargs_in, **kwargs_out
    )

    assert_allclose(output_ref, output_test)
    assert_allclose(footprint_ref, footprint_test)


@pytest.mark.parametrize("block_size", [None, (32, 32)])
def test_reproject_order(block_size):
    # Check that the order keyword argument has an effect. This is a regression
    # test for a bug that caused the order= keyword argument to be ignored when
    # in parallel/blocked reprojection.

    with fits.open(get_pkg_data_filename("data/galactic_2d.fits", package="reproject.tests")) as pf:
        hdu_in = pf[0]

        header_out = hdu_in.header.copy()
        header_out["CTYPE1"] = "RA---TAN"
        header_out["CTYPE2"] = "DEC--TAN"
        header_out["CRVAL1"] = 266.39311
        header_out["CRVAL2"] = -28.939779

        array_out_bilinear = reproject_interp(
            hdu_in,
            header_out,
            return_footprint=False,
            order="bilinear",
            block_size=block_size,
        )

        array_out_biquadratic = reproject_interp(
            hdu_in,
            header_out,
            return_footprint=False,
            order="biquadratic",
            block_size=block_size,
        )

        with pytest.raises(AssertionError):
            assert_allclose(array_out_bilinear, array_out_biquadratic)


def test_reproject_block_size_broadcasting():
    # Regression test for a bug that caused the default chunk size to be
    # inadequate when using broadcasting in parallel mode

    array_in = np.ones((350, 250, 150))
    wcs_in = WCS(naxis=2)
    wcs_out = WCS(naxis=2)

    reproject_interp(
        (array_in, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        parallel=1,
        return_footprint=False,
    )

    # Specifying a block size that is missing the extra dimension should work fine:

    reproject_interp(
        (array_in, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        parallel=1,
        return_footprint=False,
        block_size=(100, 100),
    )

    # Specifying a block size with the extra dimension should work provided it matches the final output shape

    reproject_interp(
        (array_in, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        parallel=1,
        return_footprint=False,
        block_size=(350, 100, 100),
    )

    # But it should fail if we specify a block size that is smaller that the total array shape

    with pytest.raises(ValueError, match="block shape should either match output data shape"):
        reproject_interp(
            (array_in, wcs_in),
            wcs_out,
            shape_out=(300, 300),
            parallel=1,
            return_footprint=False,
            block_size=(100, 100, 100),
        )


def test_reproject_dask_return_type(dask_method):
    # Regression test for a bug that caused dask arrays to not be computable
    # when using return_type='dask' when the input was a dask array.

    array_in = da.ones((35, 250, 150))
    wcs_in = WCS(naxis=2)
    wcs_out = WCS(naxis=2)

    result_numpy = reproject_interp(
        (array_in, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        return_type="numpy",
        return_footprint=False,
        dask_method=dask_method,
    )

    result_dask = reproject_interp(
        (array_in, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        block_size=(100, 100),
        return_type="dask",
        return_footprint=False,
        dask_method=dask_method,
    )

    assert_allclose(result_numpy, result_dask.compute(scheduler="synchronous"))


def test_auto_block_size(dask_method):
    # Unit test to make sure that specifying block_size='auto' works

    array_in = da.ones((350, 250, 150))
    wcs_in = WCS(naxis=2)
    wcs_out = WCS(naxis=2)

    # When block size and parallel aren't specified, can't return as dask arrays
    with pytest.raises(ValueError, match="Output cannot be returned as dask arrays"):
        reproject_interp(
            (array_in, wcs_in),
            wcs_out,
            shape_out=(300, 300),
            return_type="dask",
            dask_method=dask_method,
        )

    array_out, footprint_out = reproject_interp(
        (array_in, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        return_type="dask",
        block_size="auto",
        dask_method=dask_method,
    )

    assert array_out.chunksize[0] == 350
    assert footprint_out.chunksize[0] == 350


@pytest.mark.parametrize("itemsize", (4, 8))
def test_bigendian_dask(itemsize, dask_method):

    # Regression test for an endianness issue that occurred when the input was
    # passed in as (dask_array, wcs) and the dask array was big endian.

    array_in_le = da.ones((35, 250, 150), dtype=f">f{itemsize}")
    array_in_be = da.ones((35, 250, 150), dtype=f"<f{itemsize}")
    wcs_in = WCS(naxis=2)
    wcs_out = WCS(naxis=2)

    array_out_be, _ = reproject_interp(
        (array_in_be, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        block_size=(100, 100),
        dask_method=dask_method,
    )

    array_out_le, _ = reproject_interp(
        (array_in_le, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        block_size=(100, 100),
        dask_method=dask_method,
    )

    assert_allclose(array_out_be, array_out_le)


def test_reproject_parallel_broadcasting(caplog, dask_method):

    # Unit test for reprojecting using parallelization along broadcasted
    # dimensions

    # Ensure caplog captures INFO regardless of the global pytest log
    # configuration, since this test relies on it.
    caplog.set_level(logging.INFO, logger="reproject._common")

    array_in = np.ones((350, 250, 150))
    wcs_in = WCS(naxis=2)
    wcs_out = WCS(naxis=2)

    # By default if we give a block size that is only in the WCS dimensions,
    # the data has a single chunk in the broadcasted dimensions

    array1 = reproject_interp(
        (array_in, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        parallel=1,
        return_footprint=False,
        block_size=(100, 100),
        return_type="dask",
        dask_method=dask_method,
    )

    assert array1.chunksize == (350, 100, 100)

    assert "Broadcasting is being used" in caplog.text
    assert "Not parallelizing along broadcasted dimension" in caplog.text
    caplog.clear()

    # However, we can also have one chunk in the WCS dimensions and several in
    # the broadcasted dimensions.

    array2 = reproject_interp(
        (array_in, wcs_in),
        wcs_out,
        shape_out=(300, 300),
        parallel=1,
        return_footprint=False,
        block_size=(1, 300, 300),
        return_type="dask",
        dask_method=dask_method,
    )

    assert array2.chunksize == (1, 300, 300)

    assert "Broadcasting is being used" in caplog.text
    assert "Parallelizing along broadcasted dimension" in caplog.text
    caplog.clear()

    # However, we can also have one chunk in the WCS dimensions and several in
    # the broadcasted dimensions.

    with pytest.raises(ValueError, match="block shape should either match output"):
        reproject_interp(
            (array_in, wcs_in),
            wcs_out,
            shape_out=(300, 300),
            parallel=1,
            return_footprint=False,
            block_size=(1, 100, 100),
            return_type="dask",
            dask_method=dask_method,
        )


@pytest.mark.parametrize("order", ["nearest-neighbor", "bilinear", "biquadratic", "bicubic"])
def test_reproject_order_method(order):

    # Make sure that the dask-image and scipy map_coordinates implementations
    # end up producing the same results

    with fits.open(get_pkg_data_filename("data/galactic_2d.fits", package="reproject.tests")) as pf:

        hdu_in = pf[0]

        header_out = hdu_in.header.copy()
        header_out["CTYPE1"] = "RA---TAN"
        header_out["CTYPE2"] = "DEC--TAN"
        header_out["CRVAL1"] = 266.39311
        header_out["CRVAL2"] = -28.939779

        data = da.from_array(ArrayWrapper(hdu_in.data), name="test")

        array_out_memmap = reproject_interp(
            (data, hdu_in.header),
            header_out,
            return_footprint=False,
            order=order,
            block_size=(10, 10),
            dask_method="memmap",
        )

        array_out_none = reproject_interp(
            (data, hdu_in.header),
            header_out,
            return_footprint=False,
            order=order,
            block_size=(10, 10),
            dask_method="none",
        )

        assert_allclose(array_out_memmap, array_out_none)



from .._wcs_utils import has_celestial
from ..utils import parse_input_data, parse_output_projection
from ._core import healpix_to_image, image_to_healpix
from ._utils import parse_coord_system, parse_input_healpix_data

__all__ = ["reproject_from_healpix", "reproject_to_healpix"]


def reproject_from_healpix(
    input_data, output_projection, shape_out=None, hdu_in=1, order="bilinear", nested=None, field=0
):
    """
    Reproject data from a HEALPIX projection to a standard projection.

    Parameters
    ----------
    input_data : object
        The input data to reproject. This can be:

            * The name of a HEALPIX FITS file
            * A `~astropy.io.fits.TableHDU` or `~astropy.io.fits.BinTableHDU`
              instance
            * A tuple where the first element is a `~numpy.ndarray` and the
              second element is a `~astropy.coordinates.BaseCoordinateFrame`
              instance or a string alias for a coordinate frame.

    output_projection : `~astropy.wcs.wcsapi.BaseLowLevelWCS` or `~astropy.wcs.wcsapi.BaseHighLevelWCS` or `~astropy.io.fits.Header`
         The output projection, which can be either a
         `~astropy.wcs.wcsapi.BaseLowLevelWCS`,
         `~astropy.wcs.wcsapi.BaseHighLevelWCS`, or a `~astropy.io.fits.Header`
         instance.
    shape_out : tuple, optional
        If ``output_projection`` is a WCS instance, the shape of the output
        data should be specified separately.
    hdu_in : int or str, optional
        If ``input_data`` is a FITS file, specifies the HDU to use.
        (the default HDU for HEALPIX data is 1, unlike with image files where
        it is generally 0).
    order : int or str, optional
        The order of the interpolation (if ``mode`` is set to
        ``'interpolation'``). This can be either one of the following strings:

            * 'nearest-neighbor'
            * 'bilinear'

        or an integer. A value of ``0`` indicates nearest neighbor
        interpolation.
    nested : bool, optional
        The order of the healpix_data, either nested (True) or ring (False).
        If a FITS file is passed in, this is determined from the header.
    field : int, optional
        The column to read from the HEALPIX FITS file. If the fits file is a
        partial-sky file, field=0 corresponds to the first column after the
        pixel index column.

    Returns
    -------
    array_new : `~numpy.ndarray`
        The reprojected array.
    footprint : `~numpy.ndarray`
        Footprint of the input array in the output array. Values of 0 indicate
        no coverage or valid values in the input image, while values of 1
        indicate valid values.
    """

    array_in, coord_system_in, nested = parse_input_healpix_data(
        input_data, hdu_in=hdu_in, field=field, nested=nested
    )
    wcs_out, shape_out = parse_output_projection(output_projection, shape_out=shape_out)

    if nested is None:
        raise ValueError(
            "Could not determine whether the data follows the "
            "'ring' or 'nested' ordering, so you should set "
            "nested=True or nested=False explicitly."
        )

    return healpix_to_image(
        array_in, coord_system_in, wcs_out, shape_out, order=order, nested=nested
    )


def reproject_to_healpix(
    input_data, coord_system_out, hdu_in=0, order="bilinear", nested=False, nside=128
):
    """
    Reproject data from a standard projection to a HEALPIX projection.

    Parameters
    ----------
    input_data : object
        The input data to reproject. This can be:

            * The name of a FITS file as a `str` or a `pathlib.Path` object
            * An `~astropy.io.fits.HDUList` object
            * An image HDU object such as a `~astropy.io.fits.PrimaryHDU`,
              `~astropy.io.fits.ImageHDU`, or `~astropy.io.fits.CompImageHDU`
              instance
            * A tuple where the first element is a `~numpy.ndarray` and the
              second element is either a
              `~astropy.wcs.wcsapi.BaseLowLevelWCS`,
              `~astropy.wcs.wcsapi.BaseHighLevelWCS`, or a
              `~astropy.io.fits.Header` object
            * An `~astropy.nddata.NDData` object from which the ``.data`` and
              ``.wcs`` attributes will be used as the input data.

    coord_system_out : `~astropy.coordinates.BaseCoordinateFrame` or str
        The output coordinate system for the HEALPIX projection.
    hdu_in : int or str, optional
        If ``input_data`` is a FITS file or an `~astropy.io.fits.HDUList`
        instance, specifies the HDU to use.
    order : int or str, optional
        The order of the interpolation (if ``mode`` is set to
        ``'interpolation'``). This can be either one of the following strings:

            * 'nearest-neighbor'
            * 'bilinear'
            * 'biquadratic'
            * 'bicubic'

        or an integer. A value of ``0`` indicates nearest neighbor
        interpolation.
    nested : bool
        The order of the healpix_data, either nested (`True`) or ring (`False`).
    nside : int, optional
        The resolution of the HEALPIX projection.

    Returns
    -------
    array_new : `~numpy.ndarray`
        The reprojected array.
    footprint : `~numpy.ndarray`
        Footprint of the input array in the output array. Values of 0 indicate
        no coverage or valid values in the input image, while values of 1
        indicate valid values.
    """

    array_in, wcs_in = parse_input_data(input_data, hdu_in=hdu_in)
    coord_system_out = parse_coord_system(coord_system_out)

    if (
        has_celestial(wcs_in)
        and wcs_in.low_level_wcs.pixel_n_dim == 2
        and wcs_in.low_level_wcs.world_n_dim == 2
        and array_in.ndim == 2
    ):
        return image_to_healpix(
            array_in, wcs_in, coord_system_out, nside=nside, order=order, nested=nested
        )
    else:
        raise NotImplementedError(
            "Only data with a 2-d celestial WCS can be reprojected to a HEALPIX projection"
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import random

import numpy as np
import pytest
from astropy.io import fits
from astropy.io.fits import Header
from astropy.wcs import WCS
from numpy.testing import assert_allclose

from reproject import reproject_exact, reproject_interp
from reproject.mosaicking._coadd import reproject_and_coadd
from reproject.tests.helpers import array_footprint_to_hdulist

ATOL = 1.0e-9

DATA = os.path.join(os.path.dirname(__file__), "..", "..", "tests", "data")


@pytest.fixture(params=[reproject_interp, reproject_exact], ids=["interp", "exact"])
def reproject_function(request):
    return request.param


@pytest.fixture(params=[False, True])
def intermediate_memmap(request):
    return request.param


class TestReprojectAndCoAdd:
    def setup_method(self, method):
        self.wcs = WCS(naxis=2)
        self.wcs.wcs.ctype = "RA---TAN", "DEC--TAN"
        self.wcs.wcs.crpix = 322, 151
        self.wcs.wcs.crval = 43, 23
        self.wcs.wcs.cdelt = -0.1, 0.1
        self.wcs.wcs.equinox = 2000.0

        self.array = np.random.random((399, 334))

    def _get_tiles(self, views):
        # Given a list of views as (imin, imax, jmin, jmax), construct
        #  tiles that can be passed into the co-adding code

        input_data = []

        for view in views:
            array = self.array[view].copy()
            wcs = self.wcs.deepcopy()
            wcs = wcs[view]
            input_data.append((array, wcs))

        return input_data

    @property
    def _nonoverlapping_views(self):
        ie = (0, 122, 233, 245, 334)
        je = (0, 44, 45, 333, 335, 399)

        views = []
        for i in range(4):
            for j in range(5):
                views.append(np.s_[je[j] : je[j + 1], ie[i] : ie[i + 1]])

        return views

    @property
    def _overlapping_views(self):
        ie = (0, 122, 233, 245, 334)
        je = (0, 44, 45, 333, 335, 399)

        views = []
        for i in range(4):
            for j in range(5):
                views.append(np.s_[je[j] : je[j + 1] + 10, ie[i] : ie[i + 1] + 10])

        return views

    @pytest.mark.parametrize("combine_function", ["mean", "sum", "first", "last"])
    def test_coadd_no_overlap(self, combine_function, reproject_function, intermediate_memmap):
        # Make sure that if all tiles are exactly non-overlapping, and
        # we use 'sum' or 'mean', we get the exact input array back.

        input_data = self._get_tiles(self._nonoverlapping_views)

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function=combine_function,
            reproject_function=reproject_function,
        )

        assert_allclose(array, self.array, atol=ATOL)
        assert_allclose(footprint, 1, atol=ATOL)

    def test_coadd_with_overlap(self, reproject_function, intermediate_memmap):
        # Here we make the input tiles overlapping. We can only check the
        # mean, not the sum.

        input_data = self._get_tiles(self._overlapping_views)

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function="mean",
            reproject_function=reproject_function,
        )

        assert_allclose(array, self.array, atol=ATOL)

    def test_coadd_with_outputs(self, tmp_path, reproject_function, intermediate_memmap):
        # Test the options to specify output array/footprint

        input_data = self._get_tiles(self._overlapping_views)

        output_array = np.memmap(
            tmp_path / "array.np", mode="w+", dtype=float, shape=self.array.shape
        )
        output_footprint = np.memmap(
            tmp_path / "footprint.np", mode="w+", dtype=float, shape=self.array.shape
        )

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function="mean",
            reproject_function=reproject_function,
            output_array=output_array,
            output_footprint=output_footprint,
        )

        assert_allclose(output_array, self.array, atol=ATOL)
        assert_allclose(output_footprint, footprint, atol=ATOL)

    @pytest.mark.parametrize("combine_function", ["first", "last", "min", "max"])
    def test_coadd_with_overlap_first_last(self, reproject_function, combine_function):
        views = self._overlapping_views
        input_data = self._get_tiles(views)

        # Make each of the overlapping tiles different
        for i, (array, wcs) in enumerate(input_data):
            # We give each tile integer values that range from 0 to 19 but we
            # deliberately don't make the first one 0 and the last one 19 so
            # that min/max differs from first/last.
            input_data[i] = (np.full_like(array, (i + 7) % 20), wcs)

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function=combine_function,
            reproject_function=reproject_function,
        )

        # Test that either the correct tile sets the output value in the overlap regions
        test_sequence = list(enumerate(views))

        if combine_function == "last":
            test_sequence = test_sequence[::-1]
        elif combine_function == "min":
            test_sequence = test_sequence[13:] + test_sequence[:13]
        elif combine_function == "max":
            test_sequence = (test_sequence[13:] + test_sequence[:13])[::-1]

        for i, view in test_sequence:
            # Each tile in test_sequence should overwrite the following tiles
            # in the overlap regions. We'll use NaNs to mark pixels in the
            # output array that have already been set by a preceding tile, so
            # we'll go through, check that each tile matches the non-nan pixels
            # in its region, and then set that whole region to nan.
            output_tile = array[view]
            output_values = output_tile[np.isfinite(output_tile)]
            assert_allclose(output_values, (i + 7) % 20)
            array[view] = np.nan

    def test_coadd_background_matching(self, reproject_function, intermediate_memmap):
        # Test out the background matching

        input_data = self._get_tiles(self._overlapping_views)

        for array, _ in input_data:
            array += random.uniform(-3, 3)

        # First check that without background matching the arrays don't match

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function="mean",
            reproject_function=reproject_function,
        )

        assert not np.allclose(array, self.array, atol=ATOL)

        # Now check that once the backgrounds are matched the values agree

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function="mean",
            reproject_function=reproject_function,
            match_background=True,
        )

        # The absolute values of the two arrays will be offset since any
        # solution that reproduces the offsets between images is valid

        assert_allclose(array - np.mean(array), self.array - np.mean(self.array), atol=ATOL)

    def test_coadd_background_matching_one_array(self, reproject_function, intermediate_memmap):
        # Test that background matching doesn't affect the output when there's
        # only one input image.

        input_data = [(self.array, self.wcs)]

        array_matched, footprint_matched = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function="mean",
            reproject_function=reproject_function,
            match_background=True,
        )

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function="mean",
            reproject_function=reproject_function,
            match_background=False,
        )
        np.testing.assert_allclose(array, array_matched)
        np.testing.assert_allclose(footprint, footprint_matched)

    @pytest.mark.parametrize("combine_function", ["first", "last", "min", "max", "sum", "mean"])
    @pytest.mark.parametrize("match_background", [True, False])
    def test_footprint_correct(self, reproject_function, combine_function, match_background):
        # Test that the output array is zero outside the returned footprint
        # We're running this test over a somewhat large grid of parameters, so
        # cut down the array size to avoid increasing the total test runtime
        # too much.
        slice = np.s_[::3, ::3]
        wcs1 = self.wcs[slice]
        wcs2 = self.wcs.deepcopy()
        # Add a 45-degree rotation
        wcs2.wcs.pc = np.array([[0.5, -0.5], [0.5, 0.5]])

        wcs_out = wcs1.deepcopy()
        # Expand the output WCS to go beyond the input images
        wcs_out.wcs.cdelt = 2 * wcs1.wcs.cdelt[0], 2 * wcs1.wcs.cdelt[1]

        # Ensure the input data is fully non-zero, so we can tell where data
        # got projected to in the output image.
        array1 = np.full_like(self.array[slice], 2)
        array2 = array1 + 5

        array, footprint = reproject_and_coadd(
            [(array1, wcs1), (array2, wcs2)],
            wcs_out,
            shape_out=self.array.shape,
            combine_function=combine_function,
            reproject_function=reproject_function,
            match_background=match_background,
        )

        assert np.all((array != 0) == (footprint > 0))

    def test_coadd_background_matching_with_nan(self, reproject_function, intermediate_memmap):
        # Test out the background matching when NaN values are present. We do
        # this by using three arrays with the same footprint but with different
        # parts masked.

        array1 = self.array.copy() + random.uniform(-3, 3)
        array2 = self.array.copy() + random.uniform(-3, 3)
        array3 = self.array.copy() + random.uniform(-3, 3)

        array1[:, 122:] = np.nan
        array2[:, :50] = np.nan
        array2[:, 266:] = np.nan
        array3[:, :199] = np.nan

        input_data = [(array1, self.wcs), (array2, self.wcs), (array3, self.wcs)]

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function="mean",
            reproject_function=reproject_function,
            match_background=True,
        )

        # The absolute values of the two arrays will be offset since any
        # solution that reproduces the offsets between images is valid

        assert_allclose(array - np.mean(array), self.array - np.mean(self.array), atol=ATOL)

    @pytest.mark.filterwarnings("ignore:unclosed file:ResourceWarning")
    @pytest.mark.parametrize("mode", ["arrays", "filenames", "hdus", "hdulist"])
    def test_coadd_with_weights(self, tmpdir, reproject_function, mode, intermediate_memmap):
        # Make sure that things work properly when specifying weights

        array1 = self.array + 1
        array2 = self.array - 1

        weight1 = np.cumsum(np.ones_like(self.array), axis=1) - 1
        weight2 = weight1[:, ::-1]

        input_data = [(array1, self.wcs), (array2, self.wcs)]

        if mode == "arrays":
            input_weights = [weight1, weight2]
        elif mode == "filenames":
            filename1 = tmpdir.join("weight1.fits").strpath
            filename2 = tmpdir.join("weight2.fits").strpath
            fits.writeto(filename1, weight1)
            fits.writeto(filename2, weight2)
            input_weights = [filename1, filename2]
        elif mode == "hdus":
            hdu1 = fits.ImageHDU(weight1)
            hdu2 = fits.ImageHDU(weight2)
            input_weights = [hdu1, hdu2]
        elif mode == "hdulist":
            hdu1 = fits.HDUList([fits.ImageHDU(weight1, header=self.wcs.to_header())])
            hdu2 = fits.HDUList([fits.ImageHDU(weight2, header=self.wcs.to_header())])
            input_weights = [hdu1, hdu2]

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function="mean",
            input_weights=input_weights,
            reproject_function=reproject_function,
            match_background=False,
        )

        expected = self.array + (2 * (weight1 / weight1.max()) - 1)

        assert_allclose(array, expected, atol=ATOL)

    @pytest.mark.filterwarnings("ignore:unclosed file:ResourceWarning")
    def test_coadd_with_weights_with_wcs(self, tmpdir, reproject_function, intermediate_memmap):
        # Make sure that things work properly when specifying weights that have offset WCS

        array1 = self.array + 1
        array2 = self.array - 1

        weight1 = np.cumsum(np.ones_like(self.array), axis=1) - 1
        weight2 = weight1[:, ::-1]

        input_data = [(array1, self.wcs), (array2, self.wcs)]

        # make weight WCS pixel scale bigger so that weights encompass data
        weightwcs = self.wcs.copy()
        weightwcs.wcs.cdelt *= 1.1

        hdu1 = fits.ImageHDU(weight1, header=weightwcs.to_header())
        hdu2 = fits.ImageHDU(weight2, header=weightwcs.to_header())
        input_weights = [hdu1, hdu2]

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=self.array.shape,
            combine_function="mean",
            input_weights=input_weights,
            reproject_function=reproject_function,
            match_background=False,
        )

        weights1_reprojected = reproject_function(
            hdu1, self.wcs, shape_out=self.array.shape, return_footprint=False
        )
        weights2_reprojected = reproject_function(
            hdu2, self.wcs, shape_out=self.array.shape, return_footprint=False
        )
        array1_reprojected = reproject_function(
            input_data[0], self.wcs, shape_out=self.array.shape, return_footprint=False
        )
        array2_reprojected = reproject_function(
            input_data[1], self.wcs, shape_out=self.array.shape, return_footprint=False
        )
        expected = (
            array1_reprojected * weights1_reprojected + array2_reprojected * weights2_reprojected
        ) / (weights1_reprojected + weights2_reprojected)

        assert_allclose(array, expected, atol=ATOL)

    @pytest.mark.parametrize("block_size_mode", (None, "block_size", "block_sizes"))
    def test_coadd_with_broadcasting(
        self, reproject_function, intermediate_memmap, block_size_mode
    ):

        # Coadding should work with broadcasting, i.e. the fact the
        # input/output WCS might have fewer dimensions than the data.

        input_data = self._get_tiles(self._overlapping_views)

        input_data = [
            (np.broadcast_to(array.reshape((1,) + array.shape), (3,) + array.shape), wcs)
            for array, wcs in input_data
        ]

        if block_size_mode == "block_size":
            kwargs = {"block_size": (1,) + self.array.shape}
        elif block_size_mode == "block_size":
            kwargs = {"block_sizes": (1,) + self.array.shape}
        else:
            kwargs = {}

        array, footprint = reproject_and_coadd(
            input_data,
            self.wcs,
            shape_out=(3,) + self.array.shape,
            combine_function="mean",
            reproject_function=reproject_function,
            **kwargs,
        )

        for index in range(3):
            assert_allclose(array[index], self.array, atol=ATOL)


HEADER_SOLAR_OUT = """
WCSAXES =                    2
CRPIX1  =                 90.5
CRPIX2  =                 45.5
CDELT1  =                    2
CDELT2  =                    2
CUNIT1  = 'deg'
CUNIT2  = 'deg'
CTYPE1  = 'HGLN-CAR'
CTYPE2  = 'HGLT-CAR'
CRVAL1  =                  0.0
CRVAL2  =                  0.0
LONPOLE =                  0.0
LATPOLE =                 90.0
DATE-OBS= '2011-02-15T00:14:03.654'
MJD-OBS =      55607.009764514
MJD-OBS =      55607.009764514
"""


@pytest.mark.array_compare()
def test_coadd_solar_map():
    # This is a test that exercises a lot of different parts of the mosaicking
    # code. The idea is to take three solar images from different viewpoints
    # and combine them into a single one. This uses weight maps that are not
    # uniform and also include NaN values.

    pytest.importorskip("sunpy", minversion="6.0.1")
    from sunpy.map import Map, all_coordinates_from_map

    # Load in three images from different viewpoints around the Sun
    filenames = ["secchi_l0_a.fits", "aia_171_level1.fits", "secchi_l0_b.fits"]
    maps = [Map(os.path.join(DATA, f)) for f in filenames]

    # Produce weight maps that are centered on the solar disk and go to zero at the edges
    coordinates = tuple(map(all_coordinates_from_map, maps))
    input_weights = [coord.transform_to("heliocentric").z.value for coord in coordinates]
    input_weights = [(w / np.nanmax(w)) ** 4 for w in input_weights]

    shape_out = [90, 180]
    wcs_out = WCS(Header.fromstring(HEADER_SOLAR_OUT, sep="\n"))
    scales = [1 / 6, 1, 1 / 6]

    input_data = tuple((a.data * scale, a.wcs) for (a, scale) in zip(maps, scales, strict=True))

    array, footprint = reproject_and_coadd(
        input_data,
        wcs_out,
        shape_out,
        input_weights=input_weights,
        reproject_function=reproject_interp,
        match_background=True,
    )

    header_out = wcs_out.to_header()

    return array_footprint_to_hdulist(array, footprint, header_out)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Astropy affiliated package for image reprojection (resampling).
"""

from .adaptive import reproject_adaptive  # noqa
from .healpix import reproject_from_healpix, reproject_to_healpix  # noqa
from .interpolation import reproject_interp  # noqa
from .spherical_intersect import reproject_exact  # noqa
from .version import __version__  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS

from ...interpolation.tests.test_core import as_high_level_wcs
from .._core import _reproject_celestial


def test_reproject_celestial_slices_2d():
    header_in = fits.Header.fromtextfile(get_pkg_data_filename("../../tests/data/gc_ga.hdr"))
    header_out = fits.Header.fromtextfile(get_pkg_data_filename("../../tests/data/gc_eq.hdr"))

    array_in = np.ones((100, 100))

    wcs_in = WCS(header_in)
    wcs_out = WCS(header_out)

    _reproject_celestial(array_in, wcs_in, wcs_out, (200, 200))


DATA = np.array([[1, 2], [3, 4]], dtype=np.int64)

INPUT_HDR = """
WCSAXES =                    2 / Number of coordinate axes
CRPIX1  =              299.628 / Pixel coordinate of reference point
CRPIX2  =              299.394 / Pixel coordinate of reference point
CDELT1  =         -0.001666666 / [deg] Coordinate increment at reference point
CDELT2  =          0.001666666 / [deg] Coordinate increment at reference point
CUNIT1  = 'deg'                / Units of coordinate increment and value
CUNIT2  = 'deg'                / Units of coordinate increment and value
CTYPE1  = 'GLON-CAR'           / galactic longitude, plate caree projection
CTYPE2  = 'GLAT-CAR'           / galactic latitude, plate caree projection
CRVAL1  =                  0.0 / [deg] Coordinate value at reference point
CRVAL2  =                  0.0 / [deg] Coordinate value at reference point
LONPOLE =                  0.0 / [deg] Native longitude of celestial pole
LATPOLE =                 90.0 / [deg] Native latitude of celestial pole
"""

OUTPUT_HDR = """
WCSAXES =                    2 / Number of coordinate axes
CRPIX1  =                  2.5 / Pixel coordinate of reference point
CRPIX2  =                  2.5 / Pixel coordinate of reference point
CDELT1  =         -0.001500000 / [deg] Coordinate increment at reference point
CDELT2  =          0.001500000 / [deg] Coordinate increment at reference point
CUNIT1  = 'deg'                / Units of coordinate increment and value
CUNIT2  = 'deg'                / Units of coordinate increment and value
CTYPE1  = 'RA---TAN'           / Right ascension, gnomonic projection
CTYPE2  = 'DEC--TAN'           / Declination, gnomonic projection
CRVAL1  =        267.183880241 / [deg] Coordinate value at reference point
CRVAL2  =        -28.768527143 / [deg] Coordinate value at reference point
LONPOLE =                180.0 / [deg] Native longitude of celestial pole
LATPOLE =        -28.768527143 / [deg] Native latitude of celestial pole
EQUINOX =               2000.0 / [yr] Equinox of equatorial coordinates
"""

MONTAGE_REF = np.array(
    [
        [np.nan, 2.0, 2.0, np.nan],
        [1.0, 1.6768244, 3.35364754, 4.0],
        [1.0, 1.6461656, 3.32308315, 4.0],
        [np.nan, 3.0, 3.0, np.nan],
    ]
)

# Same geometry as INPUT_HDR/OUTPUT_HDR but entirely in ICRS (no Galactic conversion).
# This isolates the spherical polygon intersection accuracy from any differences
# in Galactic coordinate conversions between astropy and Montage.

INPUT_HDR_ICRS = """
WCSAXES =                    2 / Number of coordinate axes
CRPIX1  =              299.628 / Pixel coordinate of reference point
CRPIX2  =              299.394 / Pixel coordinate of reference point
CDELT1  =         -0.001666666 / [deg] Coordinate increment at reference point
CDELT2  =          0.001666666 / [deg] Coordinate increment at reference point
CUNIT1  = 'deg'                / Units of coordinate increment and value
CUNIT2  = 'deg'                / Units of coordinate increment and value
CTYPE1  = 'RA---CAR'           / Right ascension, plate caree projection
CTYPE2  = 'DEC--CAR'           / Declination, plate caree projection
CRVAL1  =                  0.0 / [deg] Coordinate value at reference point
CRVAL2  =                  0.0 / [deg] Coordinate value at reference point
LONPOLE =                180.0 / [deg] Native longitude of celestial pole
LATPOLE =                  0.0 / [deg] Native latitude of celestial pole
EQUINOX =               2000.0 / [yr] Equinox of equatorial coordinates
"""

OUTPUT_HDR_ICRS = """
WCSAXES =                    2 / Number of coordinate axes
CRPIX1  =                  2.5 / Pixel coordinate of reference point
CRPIX2  =                  2.5 / Pixel coordinate of reference point
CDELT1  =         -0.001500000 / [deg] Coordinate increment at reference point
CDELT2  =          0.001500000 / [deg] Coordinate increment at reference point
CUNIT1  = 'deg'                / Units of coordinate increment and value
CUNIT2  = 'deg'                / Units of coordinate increment and value
CTYPE1  = 'RA---TAN'           / Right ascension, gnomonic projection
CTYPE2  = 'DEC--TAN'           / Declination, gnomonic projection
CRVAL1  =      0.4960464682000 / [deg] Coordinate value at reference point
CRVAL2  =     -0.4956564684000 / [deg] Coordinate value at reference point
LONPOLE =                180.0 / [deg] Native longitude of celestial pole
LATPOLE =     -0.4956564684000 / [deg] Native latitude of celestial pole
EQUINOX =               2000.0 / [yr] Equinox of equatorial coordinates
"""

# Reference values from Montage mProject (run with -f -x 1.0 flags)
MONTAGE_ICRS_REF = np.array(
    [
        [1.0, 1.5555337416634756, 2.0, np.nan],
        [2.111112839105374, 2.6666450687397223, 3.1111069641586675, np.nan],
        [3.0, 3.5555337849185467, 4.0, np.nan],
        [np.nan, np.nan, np.nan, np.nan],
    ]
)


@pytest.mark.parametrize("wcsapi", (False, True))
def test_reproject_celestial_montage(wcsapi):
    # Accuracy compared to Montage

    wcs_in = WCS(fits.Header.fromstring(INPUT_HDR, sep="\n"))
    wcs_out = WCS(fits.Header.fromstring(OUTPUT_HDR, sep="\n"))

    if wcsapi:  # Enforce a pure wcsapi API
        wcs_in, wcs_out = as_high_level_wcs(wcs_in), as_high_level_wcs(wcs_out)

    array, footprint = _reproject_celestial(DATA, wcs_in, wcs_out, (4, 4))

    # The ~10% disagreement with Montage is due to differences in the
    # Galactic-to-ICRS coordinate transformation between astropy and Montage,
    # not the polygon intersection algorithm (see test_reproject_celestial_icrs_montage
    # which agrees to ~1e-6 when no Galactic conversion is involved).
    np.testing.assert_allclose(array, MONTAGE_REF, rtol=0.09)


@pytest.mark.parametrize("wcsapi", (False, True))
def test_reproject_celestial_icrs_montage(wcsapi):
    # Accuracy compared to Montage, using only ICRS coordinates (no Galactic
    # conversion). This isolates the polygon intersection algorithm from
    # any differences in Galactic-to-ICRS coordinate transformations between
    # astropy and Montage.

    wcs_in = WCS(fits.Header.fromstring(INPUT_HDR_ICRS, sep="\n"))
    wcs_out = WCS(fits.Header.fromstring(OUTPUT_HDR_ICRS, sep="\n"))

    if wcsapi:
        wcs_in, wcs_out = as_high_level_wcs(wcs_in), as_high_level_wcs(wcs_out)

    array, footprint = _reproject_celestial(DATA, wcs_in, wcs_out, (4, 4))

    np.testing.assert_allclose(array, MONTAGE_ICRS_REF, rtol=1.0e-5)


def test_reproject_flipping():
    # Regression test for a bug that caused issues when the WCS was oriented
    # in a way that meant polygon vertices were clockwise.

    wcs_in = WCS(fits.Header.fromstring(INPUT_HDR, sep="\n"))
    wcs_out = WCS(fits.Header.fromstring(OUTPUT_HDR, sep="\n"))
    array1, footprint1 = _reproject_celestial(DATA, wcs_in, wcs_out, (4, 4))

    # Repeat with an input that is flipped horizontally with the equivalent WCS
    wcs_in_flipped = WCS(fits.Header.fromstring(INPUT_HDR, sep="\n"))
    wcs_in_flipped.wcs.cdelt[0] = -wcs_in_flipped.wcs.cdelt[0]
    wcs_in_flipped.wcs.crpix[0] = 3 - wcs_in_flipped.wcs.crpix[0]
    array2, footprint2 = _reproject_celestial(DATA[:, ::-1], wcs_in_flipped, wcs_out, (4, 4))

    # Repeat with an output that is flipped horizontally with the equivalent WCS
    wcs_out_flipped = WCS(fits.Header.fromstring(OUTPUT_HDR, sep="\n"))
    wcs_out_flipped.wcs.cdelt[0] = -wcs_out_flipped.wcs.cdelt[0]
    wcs_out_flipped.wcs.crpix[0] = 5 - wcs_out_flipped.wcs.crpix[0]
    array3, footprint3 = _reproject_celestial(DATA, wcs_in, wcs_out_flipped, (4, 4))
    array3, footprint3 = array3[:, ::-1], footprint3[:, ::-1]

    np.testing.assert_allclose(array1, array2, rtol=1.0e-10)
    np.testing.assert_allclose(array1, array3, rtol=1.0e-10)

    np.testing.assert_allclose(footprint1, footprint2, rtol=1.0e-10)
    np.testing.assert_allclose(footprint1, footprint3, rtol=1.0e-10)



import dask.array as da
import numpy as np
import pytest
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS

from reproject._wcs_utils import has_celestial
from reproject.conftest import set_wcs_array_shape
from reproject.tests.helpers import assert_wcs_allclose
from reproject.utils import (
    hdu_to_numpy_memmap,
    parse_input_data,
    parse_input_shape,
    parse_output_projection,
)


@pytest.mark.filterwarnings("ignore:unclosed file:ResourceWarning")
def test_parse_input_data(tmpdir, valid_celestial_input_data, request):
    array_ref, wcs_ref, input_value, kwargs = valid_celestial_input_data

    data, wcs = parse_input_data(input_value, **kwargs)
    assert isinstance(data, da.Array | np.ndarray)
    np.testing.assert_allclose(data, array_ref)
    assert_wcs_allclose(wcs, wcs_ref)


def test_parse_input_data_invalid():
    data = np.ones((30, 40))

    with pytest.raises(TypeError, match="input_data should either be an HDU object"):
        parse_input_data(data)


def test_parse_input_data_missing_hdu_in():
    hdulist = fits.HDUList(
        [fits.PrimaryHDU(data=np.ones((30, 40))), fits.ImageHDU(data=np.ones((20, 30)))]
    )

    with pytest.raises(ValueError, match="More than one HDU"):
        parse_input_data(hdulist)


def test_parse_input_data_distortion_map():
    # Verify that the file can be successfully loaded and parsed
    fname = get_pkg_data_filename("data/image_with_distortion_map.fits", package="reproject.tests")
    parse_input_data(fname, hdu_in=0)


@pytest.mark.filterwarnings("ignore:unclosed file:ResourceWarning")
def test_parse_input_shape(tmpdir, valid_celestial_input_shapes):
    """
    This should support everything that parse_input_data does, *plus* an
    "array-like" argument that is just a shape rather than a populated array.
    """

    array_ref, wcs_ref, input_value, kwargs = valid_celestial_input_shapes

    shape, wcs = parse_input_shape(input_value, **kwargs)
    assert shape == array_ref.shape
    assert_wcs_allclose(wcs, wcs_ref)


def test_parse_input_shape_invalid():
    data = np.ones((30, 40))

    # Invalid
    with pytest.raises(TypeError) as exc:
        parse_input_shape(data)
    assert exc.value.args[0] == (
        "input_shape should either be an HDU object or a tuple "
        "of (array-or-shape, WCS) or (array-or-shape, Header)"
    )


def test_parse_input_shape_missing_hdu_in():
    hdulist = fits.HDUList(
        [fits.PrimaryHDU(data=np.ones((30, 40))), fits.ImageHDU(data=np.ones((20, 30)))]
    )

    with pytest.raises(ValueError) as exc:
        shape, coordinate_system = parse_input_shape(hdulist)
    assert exc.value.args[0] == (
        "More than one HDU is present, please specify HDU to use with ``hdu_in=`` option"
    )


def test_parse_output_projection(valid_celestial_output_projections):
    wcs_ref, shape_ref, output_value, kwargs = valid_celestial_output_projections

    wcs, shape = parse_output_projection(output_value, **kwargs)

    assert shape == shape_ref
    assert_wcs_allclose(wcs, wcs_ref)


def test_parse_output_projection_invalid_header(simple_celestial_fits_wcs):
    with pytest.raises(ValueError, match="Need to specify shape"):
        parse_output_projection(simple_celestial_fits_wcs.to_header())


def test_parse_output_projection_invalid_wcs(simple_celestial_fits_wcs):
    with pytest.raises(ValueError, match="Need to specify shape"):
        parse_output_projection(simple_celestial_fits_wcs)


def test_parse_output_projection_override_shape_out(simple_celestial_wcs):
    # Regression test for a bug that caused shape_out to be ignored if the
    # WCS object had array_shape set - but shape_out should override the WCS
    # shape.

    wcs_ref = simple_celestial_wcs

    set_wcs_array_shape(wcs_ref, (10, 20))

    if hasattr(wcs_ref, "low_level_wcs"):
        assert wcs_ref.low_level_wcs.array_shape == (10, 20)
    else:
        assert wcs_ref.array_shape == (10, 20)

    wcs, shape = parse_output_projection(wcs_ref, shape_out=(30, 40))

    assert shape == (30, 40)
    assert_wcs_allclose(wcs, wcs_ref)


@pytest.mark.filterwarnings("ignore::astropy.utils.exceptions.AstropyUserWarning")
@pytest.mark.filterwarnings("ignore::astropy.wcs.wcs.FITSFixedWarning")
def test_has_celestial():
    from .test_high_level import INPUT_HDR

    hdr = fits.Header.fromstring(INPUT_HDR)
    ww = WCS(hdr)
    assert ww.has_celestial
    assert has_celestial(ww)

    from astropy.wcs.wcsapi import HighLevelWCSWrapper, SlicedLowLevelWCS

    wwh = HighLevelWCSWrapper(SlicedLowLevelWCS(ww, Ellipsis))
    assert has_celestial(wwh)

    wwh2 = HighLevelWCSWrapper(SlicedLowLevelWCS(ww, [slice(0, 1), slice(0, 1)]))
    assert has_celestial(wwh2)


class TestHDUToMemmap:

    def test_compressed(self, tmp_path):

        hdu = fits.CompImageHDU(data=np.random.random((128, 128)))
        hdu.writeto(tmp_path / "test.fits")

        mmap = hdu_to_numpy_memmap(hdu)

        np.testing.assert_allclose(hdu.data, mmap)



#cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, language_level=3str
#distutils: define_macros=NPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION

# Cython implementation of the image resampling method described in "On
# resampling of Solar Images", C.E. DeForest, Solar Physics 2004

# Original version copyright (c) 2014, Ruben De Visscher. All rights reserved.
# v2 updates copyright (c) 2022, Sam Van Kooten. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np

cimport cython
cimport numpy as np
from cython cimport floating
from libc.math cimport atan2, ceil, cos, exp, fabs, floor, round, sin, sqrt
from libc.stdlib cimport qsort

import sys

np.import_array()

cdef double pi = np.pi
cdef double nan = np.nan

cdef extern from "math.h":
    int isnan(double x) nogil
    int isinf(double x) nogil


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void svd2x2_decompose(double[:,:] M, double[:,:] U, double[:] s, double[:,:] V) noexcept nogil:
    cdef double E = (M[0,0] + M[1,1]) / 2
    cdef double F = (M[0,0] - M[1,1]) / 2
    cdef double G = (M[1,0] + M[0,1]) / 2
    cdef double H = (M[1,0] - M[0,1]) / 2
    cdef double Q = sqrt(E*E + H*H)
    cdef double R = sqrt(F*F + G*G)
    s[0] = Q + R
    s[1] = Q - R
    cdef double a1 = atan2(G,F)
    cdef double a2 = atan2(H,E)
    cdef double theta = (a2 - a1) / 2
    cdef double phi = (a2 + a1) / 2
    U[0,0] = cos(phi)
    U[0,1] = -sin(phi)
    U[1,0] = sin(phi)
    U[1,1] = cos(phi)
    V[0,0] = cos(theta)
    V[0,1] = sin(theta)
    V[1,0] = -sin(theta)
    V[1,1] = cos(theta)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef double det2x2(double[:,:] M) noexcept nogil:
    return M[0,0]*M[1,1] - M[0,1]*M[1,0]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void svd2x2_compose(double[:,:] U, double[:] s, double[:,:] V, double[:,:] M) noexcept nogil:
    cdef double tmp00, tmp01, tmp10, tmp11
    tmp00 = U[0,0] * s[0]
    tmp01 = U[0,1] * s[1]
    tmp10 = U[1,0] * s[0]
    tmp11 = U[1,1] * s[1]
    # Multiply with transpose of V
    M[0,0] = tmp00 * V[0,0] + tmp01 * V[0,1]
    M[0,1] = tmp00 * V[1,0] + tmp01 * V[1,1]
    M[1,0] = tmp10 * V[0,0] + tmp11 * V[0,1]
    M[1,1] = tmp10 * V[1,0] + tmp11 * V[1,1]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef double hanning_filter(double x, double y) noexcept nogil:
    x = fabs(x)
    y = fabs(y)
    if x >= 1 or y >= 1:
        return 0
    return (cos(x * pi)+1.0) * (cos(y * pi)+1.0)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef double gaussian_filter(double x, double y, double width) noexcept nogil:
    return exp(-(x*x+y*y) / (width*width) * 2)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef double clip(double x, double vmin, double vmax, int cyclic,
        int out_of_range_nearest) noexcept nogil:
    """Applies boundary conditions to an intended array coordinate.

    Specifically, if the point is outside the array bounds, this function wraps
    the coordinate if the boundary is periodic, or clamps to the nearest valid
    coordinate if desired, or else returns NaN.
    """
    if x < vmin:
        if cyclic:
            while x < vmin:
                x += (vmax-vmin)+1
        elif out_of_range_nearest:
            return vmin
        else:
            return nan
    elif x > vmax:
        if cyclic:
            while x > vmax:
                x -= (vmax-vmin)+1
        elif out_of_range_nearest:
            return vmax
        else:
            return nan
    return x


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef bint sample_array(const floating[:,:,:] source, floating[:] dest,
        double x, double y, int x_cyclic, int y_cyclic,
        bint out_of_range_nearest) noexcept nogil:
    x = clip(x, 0, source.shape[2] - 1, x_cyclic, out_of_range_nearest)
    y = clip(y, 0, source.shape[1] - 1, y_cyclic, out_of_range_nearest)

    if isnan(x) or isnan(y):
        # Indicates the coordinate is outside the array's bounds and the
        # boundary-handling mode doesn't provide an alternative coordinate.
        return False

    # Cython doesn't like a return type of (double[:], bint), so we put the
    # input data into the provided output array
    dest[:] = source[:, <int> y, <int> x]
    return True


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void calculate_jacobian(double[:, :] Ji, int center_jacobian,
        int yi, int xi,
        double[:, :, :] offset_source_x, double[:, :, :] offset_source_y,
        double[:, :, :] Jx, double[:, :, :] Jy) noexcept nogil:
    """ Utility function to calculate the Jacobian at one (yi, xi) location"""
    if center_jacobian:
        # Compute the Jacobian for the transformation applied to
        # this pixel, as finite differences.
        Ji[0,0] = -offset_source_x[yi, xi, 0] + offset_source_x[yi, xi+1, 0]
        Ji[1,0] = -offset_source_x[yi, xi, 1] + offset_source_x[yi, xi+1, 1]
        Ji[0,1] = -offset_source_y[yi, xi, 0] + offset_source_y[yi+1, xi, 0]
        Ji[1,1] = -offset_source_y[yi, xi, 1] + offset_source_y[yi+1, xi, 1]
    else:
        # Compute the Jacobian for the transformation applied to
        # this pixel, as a mean of the Jacobian a half-pixel
        # forwards and backwards.
        Ji[0,0] = (Jx[yi, xi, 0] + Jx[yi, xi+1, 0]) / 2
        Ji[1,0] = (Jx[yi, xi, 1] + Jx[yi, xi+1, 1]) / 2
        Ji[0,1] = (Jy[yi, xi, 0] + Jy[yi+1, xi, 0]) / 2
        Ji[1,1] = (Jy[yi, xi, 1] + Jy[yi+1, xi, 1]) / 2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef int cmp_func(const void* a, const void* b) noexcept nogil:
    cdef double a_v = (<double*>a)[0]
    cdef double b_v = (<double*>b)[0]
    if a_v < b_v:
        return -1
    elif a_v == b_v:
        return 0
    else:
        return 1


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void sort(double[:] a) noexcept nogil:
    qsort(&a[0], a.shape[0], a.strides[0], &cmp_func)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void despike_jacobian(double[:, :, :, :] jacobian):
    """Detects and fixes pixels where the Jacobian is extremely large

    This can occur, e.g., for an all-sky map at the point where the longitude
    wraps around. In such cases, the large Jacobian is an artefact of the
    coordinates and should be eliminated.

    The spike detection uses the typical magnitude (distance from determinant)
    of the Jacobian matrix
        Jmag2 = sum_j sum_i (J_ij**2).
    The value of Jmag2 is calculated in the 3x3 neighborhood around each pixel,
    and the 25th percentile value (the third lowest value) is kept. Anywhere
        Jmag2 > Jmag2_25pct * threshold_factor
    is marked as a spike. Threshold_factor is currently hardcoded to 10. The
    Jacobian's components at spike locations are replaced with the mean of those
    from nearby non-spike locations.

    The average-magnitude-of-Jacobian method works okay because the
    typical use case is for pixel to pixel mapping (resampling data
    sets), where the overall singular value ratio is not likely to be
    large (less than, say, 30).
    """
    # Compute the magnitude of the Jacobian
    cdef double[:, :] Jmag2 = np.empty((jacobian.shape[0], jacobian.shape[1]))
    cdef int xi, yi
    for yi in range(jacobian.shape[0]):
        for xi in range(jacobian.shape[1]):
            Jmag2[yi, xi] = (
                    jacobian[yi, xi, 0, 0]**2
                    + jacobian[yi, xi, 0, 1]**2
                    + jacobian[yi, xi, 1, 0]**2
                    + jacobian[yi, xi, 1, 1]**2)

    # Cycle through and look for outliers
    cdef double percentile, thresh
    cdef int n_contributing, ymax, xmax
    ymax = jacobian.shape[0] - 2
    xmax = jacobian.shape[1] - 2
    cdef double[:] neighborhood = np.empty(9)
    with nogil:
        for yi in range(1, ymax + 1):
            for xi in range(1, xmax + 1):
                neighborhood[0:3] = Jmag2[yi-1, xi-1:xi+2]
                neighborhood[3:6] = Jmag2[yi, xi-1:xi+2]
                neighborhood[6:] = Jmag2[yi+1, xi-1:xi+2]
                # Computing the percentile through this C function is *much*
                # faster than calling np.percentile.
                sort(neighborhood)
                percentile = neighborhood[2]
                thresh = 10 * percentile
                if Jmag2[yi, xi] > thresh:
                    # This pixel is an outlier. Replace it with an average of
                    # the neighboring, non-outlier pixels
                    fill_in_jacobian(
                            jacobian, xi, xi-1, xi+1, yi, yi-1, yi+1,
                            Jmag2, thresh)

                # Check edges
                if yi == 1 and Jmag2[0, xi] > thresh:
                    fill_in_jacobian(
                            jacobian, xi, xi-1, xi+1, 0, 0, 1,
                            Jmag2, thresh)

                if yi == ymax and Jmag2[ymax+1, xi] > thresh:
                    fill_in_jacobian(
                            jacobian, xi, xi-1, xi+1, ymax+1, ymax, ymax+1,
                            Jmag2, thresh)

                if xi == 1 and Jmag2[yi, 0] > thresh:
                    fill_in_jacobian(
                            jacobian, 0, 0, 1, yi, yi-1, yi+1, Jmag2, thresh)

                if xi == xmax and Jmag2[yi, xmax+1] > thresh:
                    fill_in_jacobian(
                            jacobian, xmax+1, xmax, xmax+1, yi, yi-1, yi+1,
                            Jmag2, thresh)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void fill_in_jacobian(double[:, :, :, :] jacobian,
        int xi, int xfirst, int xlast,
        int yi, int yfirst, int ylast,
        double[:, :] Jmag2, double thresh) noexcept nogil:
    """ Utility function that replaces a spiking Jacobian pixel """
    jacobian[yi, xi] = 0
    cdef int n_contributing = 0
    for i in range(yfirst, ylast+1):
        for j in range(xfirst, xlast+1):
            if i == yi and j == xi:
                continue
            if Jmag2[i, j] < thresh:
                n_contributing += 1
                jacobian[yi, xi, 0, 0] += jacobian[i, j, 0, 0]
                jacobian[yi, xi, 0, 1] += jacobian[i, j, 0, 1]
                jacobian[yi, xi, 1, 0] += jacobian[i, j, 1, 0]
                jacobian[yi, xi, 1, 1] += jacobian[i, j, 1, 1]
    jacobian[yi, xi, 0, 0] /= n_contributing
    jacobian[yi, xi, 0, 1] /= n_contributing
    jacobian[yi, xi, 1, 0] /= n_contributing
    jacobian[yi, xi, 1, 1] /= n_contributing


KERNELS = {}
KERNELS['hann'] = 0
KERNELS['hanning'] = KERNELS['hann']
KERNELS['gaussian'] = 1

BOUNDARY_MODES = {}
BOUNDARY_MODES['strict'] = 1
BOUNDARY_MODES['constant'] = 2
BOUNDARY_MODES['grid-constant'] = 3
BOUNDARY_MODES['ignore'] = 4
BOUNDARY_MODES['ignore_threshold'] = 5
BOUNDARY_MODES['nearest'] = 6

BAD_VALUE_MODES = {}
BAD_VALUE_MODES['strict'] = 1
BAD_VALUE_MODES['constant'] = 2
BAD_VALUE_MODES['ignore'] = 3

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def map_coordinates(const floating[:,:,:] source, floating[:,:,:] target, Ci, int max_samples_width=-1,
                    int conserve_flux=False, int progress=False, int singularities_nan=False,
                    int x_cyclic=False, int y_cyclic=False, int out_of_range_nan=False,
                    bint center_jacobian=False, bint despiked_jacobian=False,
                    str kernel='gaussian', double kernel_width=1.3,
                    double sample_region_width=4, str boundary_mode="strict",
                    double boundary_fill_value=0, double boundary_ignore_threshold=0.5,
                    str bad_value_mode="strict", double bad_fill_value=0,
                    ):
    # n.b. the source and target arrays are expected to contain three
    # dimensions---the last two are the image dimensions, while the first
    # indexes multiple images with the same coordinates. The transformation is
    # computed once, and then each image is reprojected using that
    # transformation. For the single-image case, the first dimension is still
    # required and will have size 1.
    cdef int kernel_flag
    try:
        kernel_flag = KERNELS[kernel.lower()]
    except KeyError:
        raise ValueError("'kernel' must be 'Hann' or 'Gaussian'")

    cdef int boundary_flag
    try:
        boundary_flag = BOUNDARY_MODES[boundary_mode.lower()]
    except KeyError:
        raise ValueError(
                f"boundary_mode '{boundary_mode}' not recognized") from None

    cdef int bad_val_flag
    try:
        bad_val_flag = BAD_VALUE_MODES[bad_value_mode.lower()]
    except KeyError:
        raise ValueError(
                f"bad_value_mode '{bad_value_mode}' not recognized") from None

    cdef np.ndarray[np.float64_t, ndim=3] pixel_target
    cdef int delta
    if center_jacobian:
        pixel_target = np.zeros((target.shape[1], target.shape[2], 2))
        delta = 0
    else:
        # Pad by one on all four sides of the array, so we can interpolate
        # Jacobian values from both directions at all points.
        pixel_target = np.zeros((target.shape[1]+2, target.shape[2]+2, 2))
        # With this delta set, the value of pixel_target at (0,0) will really
        # be representing (-1,-1) in the output image.
        delta = -1

    cdef int yi, xi, y, x
    for yi in range(pixel_target.shape[0]):
        for xi in range(pixel_target.shape[1]):
            pixel_target[yi,xi,0] = xi + delta
            pixel_target[yi,xi,1] = yi + delta

    cdef np.ndarray[np.float64_t, ndim=3] offset_target_x
    cdef np.ndarray[np.float64_t, ndim=3] offset_target_y
    if center_jacobian:
        # Prepare arrays marking coordinates offset by a half pixel, to allow
        # for calculating centered Jacobians for each output pixel by using the
        # corresponding input coordinate at locations offset by +/- 0.5 pixels.
        offset_target_x = np.zeros((target.shape[1], target.shape[2]+1, 2))
        offset_target_y = np.zeros((target.shape[1]+1, target.shape[2], 2))
        for yi in range(target.shape[1]):
            for xi in range(target.shape[2]):
                offset_target_x[yi,xi,0] = xi - 0.5
                offset_target_x[yi,xi,1] = yi
                offset_target_y[yi,xi,0] = xi
                offset_target_y[yi,xi,1] = yi - 0.5
            offset_target_x[yi,target.shape[2],0] = target.shape[2]-1 + 0.5
            offset_target_x[yi,target.shape[2],1] = yi
        for xi in range(target.shape[2]):
            offset_target_y[target.shape[1],xi,0] = xi
            offset_target_y[target.shape[1],xi,1] = target.shape[1]-1 + 0.5

    # These source arrays store a corresponding input-image coordinate for each
    # pixel in the output image.
    cdef np.ndarray[np.float64_t, ndim=3] pixel_source = Ci(pixel_target)
    cdef double[:,:,:] offset_source_x = None
    cdef double[:,:,:] offset_source_y = None
    cdef double[:,:,:] Jx = None
    cdef double[:,:,:] Jy = None

    if center_jacobian:
        offset_source_x = Ci(offset_target_x)
        offset_source_y = Ci(offset_target_y)
    else:
        # Pre-calculate the Jacobian at each pixel location, with values
        # representing the Jacobian halfway between two grid points, and thus
        # the values of Jx at [0, 0, :] representing
        # d(input coordinate)/d(output x) at (x=-.5, y=0) in the output image,
        # and Jy at [0, 0, :] representing d(input coordinate)/d(output y) at
        # (x=0,y=-.5).
        Jx = np.empty((target.shape[1], target.shape[2] + 1, 2))
        Jy = np.empty((target.shape[1] + 1, target.shape[2], 2))
        for yi in range(target.shape[1]):
            for xi in range(target.shape[2]):
                Jx[yi, xi, 0] = -pixel_source[yi+1, xi, 0] + pixel_source[yi+1, xi+1, 0]
                Jx[yi, xi, 1] = -pixel_source[yi+1, xi, 1] + pixel_source[yi+1, xi+1, 1]
                Jy[yi, xi, 0] = -pixel_source[yi, xi+1, 0] + pixel_source[yi+1, xi+1, 0]
                Jy[yi, xi, 1] = -pixel_source[yi, xi+1, 1] + pixel_source[yi+1, xi+1, 1]
            xi = target.shape[2]
            Jx[yi, xi, 0] = -pixel_source[yi+1, xi, 0] + pixel_source[yi+1, xi+1, 0]
            Jx[yi, xi, 1] = -pixel_source[yi+1, xi, 1] + pixel_source[yi+1, xi+1, 1]
        yi = target.shape[1]
        for xi in range(target.shape[2]):
            Jy[yi, xi, 0] = -pixel_source[yi, xi+1, 0] + pixel_source[yi+1, xi+1, 0]
            Jy[yi, xi, 1] = -pixel_source[yi, xi+1, 1] + pixel_source[yi+1, xi+1, 1]

        # Now trim the padding we added earlier. Since `delta` was used above,
        # the value at (0,0) will now truly represent (0,0) and so on. After
        # this point, pixel_source is the same for both the centered and
        # uncentered Jacobian paths.
        pixel_source = pixel_source[1:-1, 1:-1]

    cdef double[:,:] Ji = np.zeros((2, 2))
    cdef double[:, :, :, :] jacobian = None
    if despiked_jacobian:
        # To do despiking, we need to have all the final Jacobian values
        # computed and ready. If we're not despiking, there's no need to hold
        # all the values in memory at once.
        jacobian = np.empty((target.shape[1], target.shape[2], 2, 2))
        for yi in range(target.shape[1]):
            for xi in range(target.shape[2]):
                calculate_jacobian(Ji, center_jacobian, yi, xi,
                        offset_source_x, offset_source_y, Jx, Jy)
                jacobian[yi, xi] = Ji
        despike_jacobian(jacobian)

    cdef double[:,:] Ji_padded = np.zeros((2, 2))
    cdef double[:,:] J = np.zeros((2, 2))
    cdef double[:,:] U = np.zeros((2, 2))
    cdef double[:] s = np.zeros((2,))
    cdef double[:,:] V = np.zeros((2, 2))
    cdef int samples_width
    cdef double[:] transformed = np.zeros((2,))
    cdef double[:] current_pixel_source = np.zeros((2,))
    cdef double[:] current_offset = np.zeros((2,))
    cdef double[:] weight_sum = np.empty(source.shape[0])
    cdef double ignored_weight_sum
    cdef double weight
    cdef floating[:] value = np.empty_like(source, shape=source.shape[0])
    cdef double[:] P1 = np.empty((2,))
    cdef double[:] P2 = np.empty((2,))
    cdef double[:] P3 = np.empty((2,))
    cdef double[:] P4 = np.empty((2,))
    cdef double top, bottom, left, right
    cdef double determinant = 0
    cdef bint has_sampled_this_row
    cdef bint sample_in_bounds
    with nogil:
        # Iterate through each pixel in the output image.
        for yi in range(target.shape[1]):
            for xi in range(target.shape[2]):
                if despiked_jacobian:
                    Ji = jacobian[yi, xi]
                else:
                    calculate_jacobian(Ji, center_jacobian, yi, xi,
                            offset_source_x, offset_source_y, Jx, Jy)
                if isnan(Ji[0,0]) or isnan(Ji[0,1]) or isnan(Ji[1,0]) or isnan(Ji[1,1]) or isnan(pixel_source[yi,xi,0]) or isnan(pixel_source[yi,xi,1]):
                    target[:,yi,xi] = nan
                    continue

                # Find and pad the singular values of the Jacobian.
                svd2x2_decompose(Ji, U, s, V)
                s[0] = max(1.0, s[0])
                s[1] = max(1.0, s[1])
                svd2x2_compose(U, s, V, Ji_padded)
                # Build J, the inverse of Ji, by using 1/s and swapping the
                # order of U and V.
                s[0] = 1.0/s[0]
                s[1] = 1.0/s[1]
                svd2x2_compose(V, s, U, J)

                # We'll need to sample some number of input image pixels to set
                # this output pixel. Later on, we'll compute weights to assign
                # to each input pixel, and they will be at or near zero outside
                # some range. Right now, we'll determine a search region within
                # the input image---a bounding box around those pixels that
                # will be assigned non-zero weights.
                #
                # We do that by defining a square region in the output plane
                # centered on the output pixel, and transforming its corners to
                # the input plane (using the local linearization of the
                # transformation). Those transformed coordinates will set our
                # bounding box.
                if kernel_flag == 0:
                    # The Hann window is zero outside +/-1, so
                    # that's how far we need to go.
                    #
                    # The Hann window width is twice the width of a pixel---it
                    # runs to the centers of the neighboring pixels, rather
                    # than the edges of those pixels. This ensures that, at
                    # every point, the sum of the overlapping Hann windows is
                    # 1, and therefore that every input-image pixel is fully
                    # distributed into some combination of output pixels (in
                    # the limit of a Jacobian that is constant across all
                    # output pixels).
                    P1[0] = - 1 * Ji_padded[0, 0] + 1 * Ji_padded[0, 1]
                    P1[1] = - 1 * Ji_padded[1, 0] + 1 * Ji_padded[1, 1]
                    P2[0] = + 1 * Ji_padded[0, 0] + 1 * Ji_padded[0, 1]
                    P2[1] = + 1 * Ji_padded[1, 0] + 1 * Ji_padded[1, 1]
                    P3[0] = - 1 * Ji_padded[0, 0] - 1 * Ji_padded[0, 1]
                    P3[1] = - 1 * Ji_padded[1, 0] - 1 * Ji_padded[1, 1]
                    P4[0] = + 1 * Ji_padded[0, 0] - 1 * Ji_padded[0, 1]
                    P4[1] = + 1 * Ji_padded[1, 0] - 1 * Ji_padded[1, 1]

                    # Find a bounding box around the transformed coordinates.
                    # (Check all four points at each step, in case a negative
                    # Jacobian value is mirroring the transformed pixel.)
                    top = max(P1[1], P2[1], P3[1], P4[1])
                    bottom = min(P1[1], P2[1], P3[1], P4[1])
                    right = max(P1[0], P2[0], P3[0], P4[0])
                    left = min(P1[0], P2[0], P3[0], P4[0])
                elif kernel_flag == 1:
                    # The Gaussian window is non-zero everywhere, but it's
                    # close to zero almost everywhere. Sampling the whole input
                    # image isn't tractable, so we truncate and sample only
                    # within a certain region.
                    # n.b. `s` currently contains the reciprocal of the
                    # singular values
                    top = sample_region_width / (2 * min(s[0], s[1]))
                    bottom = -top
                    right = top
                    left = -right
                else:
                    with gil:
                        raise ValueError("Invalid kernel type")

                if max_samples_width > 0 and max(right-left, top-bottom) > max_samples_width:
                    if singularities_nan:
                        target[:,yi,xi] = nan
                    else:
                        sample_in_bounds = sample_array(
                                source, value, current_pixel_source[0],
                                current_pixel_source[1], x_cyclic, y_cyclic,
                                out_of_range_nearest=boundary_flag == 6)
                        if sample_in_bounds:
                            for i in range(target.shape[0]):
                                if bad_val_flag != 1 and (isnan(value[i]) or isinf(value[i])):
                                    if bad_val_flag == 2:
                                        target[i,yi,xi] = bad_fill_value
                                    else:
                                        target[i,yi,xi] = nan
                                else:
                                    target[i,yi,xi] = value[i]
                        elif boundary_flag == 2 or boundary_flag == 3:
                            target[:,yi,xi] = boundary_fill_value
                        else:
                            target[:,yi,xi] = nan
                    continue

                top += pixel_source[yi,xi,1]
                bottom += pixel_source[yi,xi,1]
                right += pixel_source[yi,xi,0]
                left += pixel_source[yi,xi,0]

                # Draw these points in to the nearest input pixel
                bottom = ceil(bottom)
                top = floor(top)
                left = ceil(left)
                right = floor(right)

                # Handle the case that the sampling region extends beyond the
                # input image boundary. For 'strict' handling, we can set the
                # output pixel to NaN right away. For 'ignore' handling, we can
                # clamp the region to exclude the out-of-bounds samples. For
                # all other boundary modes, we still need to calculate weights
                # for each out-of-bounds sample, so we do nothing here.
                if not x_cyclic:
                    if right > source.shape[2] - 1:
                        if boundary_flag == 1:
                            target[:,yi,xi] = nan
                            continue
                        if boundary_flag == 4:
                            right = source.shape[2] - 1
                    if left < 0:
                        if boundary_flag == 1:
                            target[:,yi,xi] = nan
                            continue
                        if boundary_flag == 4:
                            left = 0
                if not y_cyclic:
                    if top > source.shape[1] - 1:
                        if boundary_flag == 1:
                            target[:,yi,xi] = nan
                            continue
                        if boundary_flag == 4:
                            top = source.shape[1] - 1
                    if bottom < 0:
                        if boundary_flag == 1:
                            target[:,yi,xi] = nan
                            continue
                        if boundary_flag == 4:
                            bottom = 0

                # Check whether the sampling region falls entirely outside the
                # input image. For strict boundary handling, this is already
                # handled above by the partial case. Otherwise, we fill in an
                # appropriate value and move along. For some projections, the
                # sampling region can become very large when well outside the
                # input image, and so this detection becomes an important
                # optimization.
                if (not x_cyclic and (right < 0 or left > source.shape[2] - 1)
                        or not y_cyclic
                            and (top < 0 or bottom > source.shape[1] - 1)):
                    if boundary_flag == 3:
                        target[:,yi,xi] = boundary_fill_value
                        continue
                    if (boundary_flag == 2
                            or boundary_flag == 4
                            or boundary_flag == 5):
                        target[:,yi,xi] = nan
                        continue
                    if boundary_flag == 6:
                        # Just sample one row or column so that we get all of
                        # the nearest values. Both kernels vary independently
                        # in x and y, so sampling the full region isn't needed
                        # when the sampled values are constant in x or in y.
                        if right < left:
                            right = left
                        if top < bottom:
                            top = bottom

                target[:,yi,xi] = 0
                weight_sum[:] = 0
                ignored_weight_sum = 0

                # Iterate through that bounding box in the input image.
                for y in range(<int> bottom, <int> top+1):
                    current_pixel_source[1] = y
                    current_offset[1] = current_pixel_source[1] - pixel_source[yi,xi,1]
                    has_sampled_this_row = False
                    for x in range(<int> left, <int> right+1):
                        current_pixel_source[0] = x
                        current_offset[0] = current_pixel_source[0] - pixel_source[yi,xi,0]
                        # Find the fractional position of the input location
                        # within the transformed ellipse.
                        transformed[0] = J[0,0] * current_offset[0] + J[0,1] * current_offset[1]
                        transformed[1] = J[1,0] * current_offset[0] + J[1,1] * current_offset[1]

                        # Compute an averaging weight to be assigned to this
                        # input location.
                        if kernel_flag == 0:
                            weight = hanning_filter(
                                    transformed[0], transformed[1])
                        elif kernel_flag == 1:
                            weight = gaussian_filter(
                                    transformed[0],
                                    transformed[1],
                                    kernel_width)
                        else:
                            with gil:
                                raise ValueError("Invalid kernel type")
                        if weight == 0:
                            # As we move along each row in the image, we'll
                            # first be seeing input-plane pixels that don't map
                            # back into the desired output region (i.e. they
                            # fall outside the Hanning window), then we'll see
                            # pixels that do get sampled for our output-plane
                            # pixel, and then we'll see more that don't. Once
                            # we're seeing that second group, we know we've
                            # found everything of interest in the row and can
                            # end this inner loop early. (One could be smart
                            # about skipping that first group of unused pixels,
                            # but doing so is less trivial, and skipping the
                            # second group is already only a small gain.)
                            if has_sampled_this_row:
                                break
                            continue
                        has_sampled_this_row = True

                        sample_in_bounds = sample_array(
                                source, value, current_pixel_source[0],
                                current_pixel_source[1], x_cyclic, y_cyclic,
                                out_of_range_nearest=(boundary_flag == 6))

                        if ((boundary_flag == 2 or boundary_flag == 3)
                                and not sample_in_bounds):
                            value[:] = boundary_fill_value
                            sample_in_bounds = True

                        if sample_in_bounds:
                            for i in range(target.shape[0]):
                                if bad_val_flag != 1 and (isnan(value[i]) or isinf(value[i])):
                                    if bad_val_flag == 2:
                                        value[i] = bad_fill_value
                                    else:
                                        # bad_val_flag is 3: 'ignore'
                                        continue
                                target[i,yi,xi] += weight * value[i]
                                weight_sum[i] += weight
                        else:
                            if boundary_flag == 5:
                                ignored_weight_sum += weight

                if boundary_flag == 5:
                    for i in range(target.shape[0]):
                        if (ignored_weight_sum / (ignored_weight_sum + weight_sum[i])
                                > boundary_ignore_threshold):
                            target[i,yi,xi] = nan
                if conserve_flux:
                    determinant = fabs(det2x2(Ji))
                for i in range(target.shape[0]):
                    target[i,yi,xi] /= weight_sum[i]
                    if conserve_flux:
                        target[i,yi,xi] *= determinant
            if progress:
                with gil:
                    sys.stdout.write("\r%d/%d done" % (yi+1, target.shape[1]))
                    sys.stdout.flush()
    if progress:
        sys.stdout.write("\n")
